/*
 * $Id: cia1.c,v 1.8 1994/12/12 16:59:44 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 * 
 * $Log: cia1.c,v $
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * Defines for registers
 *
 * Revision 1.7  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 * New debugger & arg parser and x128
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:13  jopi
 *  X64 version 0.2 PL 0
 * 
 * Revision 1.2  1993/06/13  08:19:36  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <time.h>

#include "macro.h"
#include "vmachine.h"
#include "extern.h"
#include "cia.h"


/* global */

BYTE    cia1[16];
BYTE    keyarr[8];

/* local functions */

static void update_cia1ta(int);
static void update_cia1tb(int);

/* local variables */

static CLOCK cia1tau;	/* time when cia 1 timer A is updated */
static CLOCK cia1tbu;	/* time when cia 1 timer B is updated */

static int cia1ta;	/* value of cia 1 timer A at last update */
static int cia1tb;	/* value of cia 1 timer B at last update */

static int tod1freq;	/* frequency of cia 1 TOD */
static int tod1rel;	/* TOD 1 difference to system clock */
static int tod1alarm;	/* TOD 1 alarm time */

static int cia1int;	/* interrupt register for cia1 */

/* ------------------------------------------------------------------------- */
/* CIA 1 */

void    store_cia1(ADDRESS addr, BYTE byte)
{
    if (debugflg)
	printf("store cia1[%x] %x\n", (int) addr, (int) byte);

    switch (addr) {
      case P_PORT_A: /* port A */
	cia1[0] = (cia1[0] & ~cia1[2]) | (byte & cia1[2]);
	break;

      case P_PORT_B: /* port B */
	cia1[1] = (cia1[1] & ~cia1[3]) | (byte & cia1[3]);
	break;

      case P_DDIR_A: /* data direction register for port A */
      case P_DDIR_B: /* data direction register for port B */
	cia1[addr] = byte;
	break;

      case P_TIMER_AL: /* timer A low */
      case P_TIMER_AH: /* timer A high */
	cia1[addr] = byte;
	update_cia1ta(1);
	break;

      case P_TIMER_BL: /* timer B low */
      case P_TIMER_BH: /* timer B high */
	cia1[addr] = byte;
	update_cia1tb(1);
	break;

#ifndef VIA6522
      case P_TOD_TEN: /* Time Of Day clock 1/10 s */
      case P_TOD_SEC: /* Time Of Day clock sec */
      case P_TOD_MIN: /* Time Of Day clock min */
      case P_TOD_HR:  /* Time Of Day clock hour */
	cia1[addr] = byte;
	tod1rel = tod1alarm = (int)time(NULL);
	break;
#endif

      case P_SERIAL:/* serial port shift register */
	cia1[addr] = byte;
	break;

      case P_INTREG:/* interrupt control register */
	if (byte & 0x80) {
	    /* set interrupts */
	    if (byte & 1) {
		update_cia1ta(0);
		int_clk[I_CIA1TA] = cia1tau + cia1ta;
	    }
	    if (byte & 2) {
		update_cia1tb(0);
		int_clk[I_CIA1TB] = cia1tbu + cia1tb;
	    }
	    if (byte & 4)
		int_clk[I_CIA1TOD] = 0;	/* tod not implemented */
	    if (byte & 8) {
		int_clk[I_CIA1SP] = 0;	/* serial data empty/full not impl. */
	    }
	    if (byte & 0x10)
		int_clk[I_CIA1FL] = 0;	/* FLAG1 not impl. */
	    cia1[addr] |= byte & 0x7f;
	} else {
	    /* clear interrupts */
	    if (byte & 1)
		int_clk[I_CIA1TA] = 0;
	    if (byte & 2)
		int_clk[I_CIA1TB] = 0;
	    if (byte & 4)
		int_clk[I_CIA1TOD] = 0;
	    if (byte & 8)
		int_clk[I_CIA1SP] = 0;
	    if (byte & 0x10)
		int_clk[I_CIA1FL] = 0;
	    cia1[addr] &= ~byte;
	}
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	break;

      case P_CONT_A: /* control register A */
	cia1[addr] = byte;

	/* bit 7 tod frequency */
	tod1freq = (byte & 0x80) ? 50 : 60;

	/* bit 6 serial port mode */

	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1ta(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */

	/* bit 0 start/stop timer */
	/* bit 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer */
	    update_cia1ta(0);
	}
	break;

      case P_CONT_B: /* control register B */
	cia1[addr] = byte;
	/* bit 7 set alarm/tod clock */
	/* bit 4 force load */
	if (byte & 0x10) {
	    update_cia1tb(1);
	}
	/* bit 3 timer run mode */
	/* bit 2 & 1 timer output to PB6 */
	/* bit 0 start/stop timer */
	/* bit 6, 5 timer count mode */
	if ((byte & 0x20) || !(byte & 1)) {
	    /* stop timer if mode = 01 or 11 */
	    update_cia1tb(0);
	}
	break;
    }
}


BYTE    read_cia1(ADDRESS addr)
{
    switch (addr) {
      case P_PORT_A: /* port A */
	{
	    int     t, i;
	    BYTE    j;
	    for (j = i = 0, t = cia1[1] & cia1[3]; i < 8; t >>= 1, i++)
		if (!(t & 1))
		    j |= keyarr[i];
	    if (debugflg)
		printf("read port A %d\n", j);
	    return (((j & ~cia1[2]) | (cia1[0] & cia1[2])) & ~joy1);
	}

      case P_PORT_B: /* port B */
	{
	    int     t, i;
	    BYTE    j;
	    /* keyarr is updated by intr() */

	    for (j = i = 0, t = cia1[0] & cia1[2]; i < 8; t >>= 1, i++)
		if (!(t & 1))
		    j |= keyarr[i];
	    j = ~j;
	    if (debugflg) {
		printf("read port B %d\n", j);
		printf("a: %x b:%x ca: %x cb: %x joy: %x\n",
		    (int) cia1[0], (int) cia1[1], (int) cia1[2], (int) cia1[3],
		    joy1);
	    }
	    return (((j & ~cia1[3]) | (cia1[1] & cia1[3])) & ~joy1);
	}

      case P_DDIR_A: /* data direction register for port A */
      case P_DDIR_B: /* data direction register for port B */
	return (cia1[addr]);

      case P_TIMER_AL: /* timer A low */
	update_cia1ta(0);
	return (cia1ta & 0xff);

      case P_TIMER_AH: /* timer A high */
	update_cia1ta(0);
	return ((cia1ta >> 8) & 0xff);

      case P_TIMER_BL: /* timer B low */
	update_cia1tb(0);
	return (cia1tb & 0xff);

      case P_TIMER_BH: /* timer B high */
	update_cia1tb(0);
	return ((cia1tb >> 8) & 0xff);

#ifndef VIA6522
      case P_TOD_TEN: /* Time Of Day clock 1/10 s */
      case P_TOD_SEC: /* Time Of Day clock sec */
      case P_TOD_MIN: /* Time Of Day clock min */
      case P_TOD_HR:  /* Time Of Day clock hour */
	return (cia1[addr]);
#endif

      case P_SERIAL:/* serial port shift register */
	return (cia1[addr]);

      case P_INTREG:/* interrupt control register */
	{
	    BYTE    t = cia1int | 8;   /****  This is ugly trick for x128 ****/
	    cia1int = 0;
	    return (t);
	}

      case P_CONT_A: /* control register A */
      case P_CONT_B: /* control register B */
	break;
    }
    return (cia1[addr]);
}


void    int_cia1ta(void)
{
    if (debugflg)
	printf("cia1 timer A interrupt\n");
    if (cia1[0xe] & 8) {
	/* one-shot mode */
	cia1ta = 0;
	cia1tau = clk;
	int_clk[I_CIA1TA] = 0;
    } else {
	/* continuous mode */
	update_cia1ta(1);
	int_clk[I_CIA1TA] = cia1tau + cia1ta;
    }
    cia1int = 0x80 | 1;

    /* count timer A underflow pulses */
    if ((cia1[15] & 0x61) == 0x41)
	if (--cia1tb <= 0)
	    int_clk[I_CIA1TB] = clk;
}


static void update_cia1ta(int force)
{
    if (force)
	cia1ta = cia1[4] + (cia1[5] << 8);
    else if ((cia1[14] & 0x21) == 1)
	cia1ta = (int)(clk - cia1tau + cia1ta) & 0xffff;
    cia1tau = clk;
}

/*
 * Timer B can run in 2 (4) modes
 * cia1[f] & 0x60 == 0x00   count system 02 pulses
 * cia1[f] & 0x60 == 0x40   count timer A underflows
 * cia1[f] & 0x60 == 0x20 | 0x60 count CNT pulses => counter stops
 */


void    int_cia1tb(void)
{
    if (debugflg)
	printf("cia1 timer B interrupt\n");

    if (cia1[0xf] & 8) {
	/* one-shot mode */
	cia1tb = 0;
	cia1tbu = clk;
	int_clk[I_CIA1TB] = 0;
    } else {
	/* continuous mode */
	update_cia1tb(1);
	if ((cia1[15] & 0x61) == 0x01)
	    int_clk[I_CIA1TB] = cia1tbu + cia1tb;
    }
    cia1int = 0x80 | 2;
}


static void update_cia1tb(int force)
{
    if (force) {
	cia1tb = cia1[4] + (cia1[5] << 8);
    } else if ((cia1[15] & 0x61) == 1)
	/* count system 02 pulses */
	cia1tb = (int)(clk - cia1tbu + cia1tb) & 0xffff;
    cia1tbu = clk;
}


int     show_keyarr(void)
{
    int     i, j;
    for (j = 0; j < 8; j++) {
	printf("%d:", j);
	for (i = 0x80; i; i >>= 1)
	    printf(" %d", keyarr[j] & i);
	printf("\n");
    }
    return (0);
}
