/*
 * $Id: drive.c,v 1.10 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file implaments the three levels of operation
 * on a floppy disk drive. They consist of:
 *	o Serial <=> IP
 *	o IP <=> FDC
 *	o FDC <=> Disk
 *
 *
 * Written by
 *   Teemu Rantanen  (tvr@cs.hut.fi)
 *
 * Patches by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Dan Miner       (dminer@nyx10.cs.du.edu)
 *   Jouko Valta     (jopi@zombie.oulu.fi)
 *   Germano Caronni (caronni@tik.ethz.ch)
 *
 *
 * $Log: drive.c,v $
 * Revision 1.10  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * New commands COPY and SCRATCH. IPE error codes.
 * Structure for passing Disk Header information.
 * New serial device control.
 * Decent Petscii <-> Ascii conversions in separate module.
 * Bugs in parsing Drive number and file type fixed.
 * The saga of Directory Linking Bug finally terminated. (:
 *
 * Revision 1.9  1995/01/26  13:51:18  jopi
 * Data handled dynamically to allow multiple drives.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * Split from "1541.c". ST port.
 *
 * Revision 1.7  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.6  1994/06/08  01:03:57  jopi
 * Patchlevel 2
 * Directory linking bug fixed. New option -t.
 *
 * Revision 1.5  1994/02/18  16:09:36  jopi
 * Hexdump routine. Header version 1.2
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * disk directory allocation fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:36:32  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:12:53  sonninen
 * *** empty log message ***
 *
 *
 */

/*
 * 1541.c -- 1541 Floppy Drive emulation.
 *  Emulator emulates 1541 reasonably well as long as 1541 is used
 *  through ROM-interface (no turboloaders supported). Also REL-file
 *  type is still missing.
 *
 * Mon Nov  8 02:33:47 MST 1993		Dan Miner
 *  Fixed do_validate() bug: the directory chain WASN'T being mapped in
 *  BAM, so on the 9th file creation, sector 18, 1 was overwriten with
 *  "new" directory information.  This bug affected format "n:filename,id"
 *  and ANY validates "V" after.  A nasty bug... :)
 *
 * Mon Nov  8 03:02:37 MST 1993		Dan Miner
 *  Fixed bug in do_format():
 *  After testing my new code, I notice that disk block 18, 1 wasn't
 *  being allocated.  I looked and 18,0 wasn't written when the automatic
 *  validate is done.  So, I manually do it in do_format().  Hopefully
 *  it will end many of the disk corrupting bugs.
 */

#ifdef __hpux
#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#endif
#endif

#ifdef __cplusplus
#ifdef __STDC__
#undef __STDC__
#endif
#define __STDC__ 0	/* kludge */
#endif


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef GEMDOS		/* Atari ST */
#include "atari.h"
#include "dirent.h"
#else
#include <string.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#endif
#include <memory.h>
#include <assert.h>


#ifndef STANDALONE_1541
#include "macro.h"
#include "extern.h"
#endif

#include "serial.h"
#include "drive.h"
#include "fs_cbm.h"
#include "fs_unix.h"
#include "file.h"
#include "proto.h"

#define SET_LO_HI(p, val) {*(p)++ = (val) & 0xff; *(p)++ = ((val)/256) & 0xff;}


static char sector_map[41] =
{
    0,
    21, 21, 21, 21, 21, 21, 21, 21, 21, 21,
    21, 21, 21, 21, 21, 21, 21, 19, 19, 19,
    19, 19, 19, 19, 18, 18, 18, 18, 18, 18,
    17, 17, 17, 17, 17,
    17, 17, 17, 17, 17		/* Tracks 36 - 40 are non-standard. */
};

static char *slot_type[] =
{
    "DEL", "SEQ", "PRG", "USR", "REL", "CBM", "DJJ", "FAB"
};

#ifdef STANDALONE_1541
extern DRIVE  *DriveData[];
#endif


#ifndef __CEXTRACT__

extern int  initialize_1541 ( int dev, int type );
extern int  find_devno ( int dev, char *name );
extern int  attach_floppy_image ( DRIVE *floppy, char *name, int mode );
extern void detach_floppy_image ( DRIVE *floppy );
static int  set_error_data(DRIVE *floppy, int flags);

extern int  open_1541 ( void *floppy, char *name, int length, int secondary );
extern int  close_1541 ( void *floppy, int secondary );
extern int  read_1541 ( void *floppy, BYTE *data, int secondary );
extern int  write_1541 ( void *floppy, BYTE data, int secondary );
extern void flush_1541 ( void *floppy, int secondary );

extern int  write_fs ( void *floppy, BYTE data, int secondary );
extern int  read_fs ( void *floppy, BYTE *data, int secondary );
extern int  open_fs ( void *floppy, char *name, int length, int secondary );
extern int  close_fs ( void *floppy, int secondary );

extern int  ip_execute ( DRIVE *floppy, BYTE *buf, int length );
static int  do_block_command ( DRIVE *floppy, char command, char *buffer );
static int  do_initialize ( DRIVE *floppy );
extern int  do_validate ( DRIVE *floppy );
static int  do_format ( DRIVE *floppy, char *name, BYTE *id, BYTE *minus );
static void floppy_error ( bufferinfo_t *p, int code, int track, int sector );
static void floppy_close_all_channels ( DRIVE * );

static int  floppy_parse_name ( char *name, int length, char *realname, int *reallength, int *readmode, int *filetype );
static int  floppy_create_directory ( DRIVE *floppy, char *name, int length, int filetype, int secondary, BYTE *outputptr );

extern int  check_track_sector( int track, int sector );
static off_t offset_from_track_and_sector ( int track, int sector );
static int  floppy_read_block ( int fd, BYTE *buf, int track, int sector );
static int  floppy_write_block (int fd, BYTE *buf, int track, int sector );
static int  do_copy ( DRIVE *floppy, char *dest, int length );

static void set_find_first_slot ( DRIVE *floppy, char *name, int length, int type );
static BYTE *find_next_slot ( DRIVE *floppy );
static void remove_slot ( DRIVE *floppy, BYTE *slot );
static int  write_sequential_buffer ( DRIVE *floppy, BYTE *slot, BYTE *buf, int length );

static int  alloc_free_sector ( BYTE *bam, int *track, int *sector );
static int  allocate_sector ( BYTE *bam, int track, int sector );
static int  free_sector ( BYTE *bam, int track, int sector );
static int  floppy_name_match ( BYTE *slot, char *name, int length, int type );

extern int  check_header ( int fd, hdrinfo *hdr );
static int  mystrncpy ( BYTE *d, BYTE *s, int n );
static void no_a0_pads ( BYTE *ptr, int l );

#endif /* __CEXTRACT__ */


int     initialize_1541(int dev, int type)
{
    DRIVE  *floppy = NULL;
    int     i;

    /*
     * Create instances of the disk drive
     */

    if (!floppy) {
	floppy = (DRIVE *)malloc(sizeof(DRIVE));
	assert(floppy);
	memset (floppy, 0, sizeof(DRIVE));	/* init all pointers */
    }

    floppy->type     = type;
    floppy->ActiveFd = -1;

    for (i = 0; i < 15; i++)
	floppy->buffers[i].mode = BUFFER_NOT_IN_USE;

    floppy->buffers[15].mode = BUFFER_COMMAND_CHANNEL;
    floppy->buffers[15].buffer = (BYTE *) malloc(256);

#ifdef STANDALONE_1541
    DriveData[dev & 3] = floppy;

#else
    /*
     * 'type' specifies what kind of emulation is selected
     */

    if (!(type & DT_FS)) {
	if (attach_serial_device(dev, (char *)floppy, "1541 Disk Drive",
	    read_1541, write_1541, open_1541, close_1541, flush_1541)) {
	    printf("could not initialize 1541 ????\n");
	    return(-1);
	}
    }
    else {
	if (attach_serial_device(dev, (char *)floppy, "1541 FS Drive",
	    read_fs, write_fs, open_fs, close_fs, NULL)) {
	    printf("could not initialize FS 1541 ????\n");
	    return(-1);
	}
    }
#endif

    floppy_error(&floppy->buffers[15], 73, 0, 0);
    return 0;
}


/*
 * This routine is used by the standalone version and ML monitor
 * in order to get around standard serial bus connection.
 * Device number is searched in filename if it's not NULL.
 * 'dev' contains the default in range 0...3 or 8...11.
 */

int  find_devno(int dev, char *name)
{
    int    num, len;
    char  *p;

    if (name && *name) {
	p = name;
	while (isspace(*p)) ++p;

#ifndef GEMDOS
	if (sscanf(p, "%d%n", &num, &len) == 1) {
	    p += len;
#else
	if (sscanf(p, "%d", &num) == 1) {
	    while (isdigit(*p)) p++;
#endif
	    while (isspace(*p)) ++p;
	    if (*p == ':')
		return (num);
	}
    }	/* name */

    return (dev);	/* Default drive */
}


/* ------------------------------------------------------------------------- */

/*
 * Serial Bus Interface
 */

int     open_1541(void *flp, char *name, int length, int secondary)
{
DRIVE *floppy = (DRIVE *)flp;
    bufferinfo_t *p = &(floppy->buffers[secondary]);
    char    realname[256];
    int     reallength;
    int     readmode;
    int     filetype;

    BYTE   *slot;	/* Current directory entry */

#ifdef DEBUG_1541
    printf("OPEN 1541: FD = %d\n", floppy->ActiveFd);
#endif

    if (!name || !*name)
	return SERIAL_ERROR;	/* Routine was called incorrectly. */


    /*
     * No floppy in drive ?
     */
    if (floppy->ActiveFd < 0) {
	printf("No disk. Status %d\n",  SERIAL_ERROR);
	return SERIAL_ERROR; }

    /*
     * Clear error flag
     */
    floppy_error(&floppy->buffers[15], 0, 0, 0);

    /*
     * If channel is command channel, name will be used as write. Return only
     * status of last write ...
     */
    if (p->mode == BUFFER_COMMAND_CHANNEL) {
	int     n;
	int     status = SERIAL_OK;

	for (n = 0; n < length; n++)
	    status = write_1541(floppy, name[n], secondary);
	return status;
    }

    /*
     * In use ?
     */
    if (p->mode != BUFFER_NOT_IN_USE) {
	floppy_error(&floppy->buffers[15], 70, 0, 0);
	return SERIAL_ERROR;
    }

    /*
     * Filemode / type
     */
    if (secondary == 1)
	readmode = 0;
    else
	readmode = 1;

    filetype = 0;

    if (floppy_parse_name(name, length, realname, &reallength,
	&readmode, &filetype) != SERIAL_OK)
	return SERIAL_ERROR;


    /*
     * Internal buffer ?
     */
    if (*name == '#') {
	p->mode = BUFFER_MEMORY_BUFFER;
	p->buffer = (BYTE *) malloc(256);
	p->bufptr = 0;
	return SERIAL_OK;
    }
    /*
     * Directory read
     */
    if (*name == '$') {
	if (!readmode)
	    return SERIAL_ERROR;

	p->mode = BUFFER_DIRECTORY_READ;
	p->buffer = (BYTE *) malloc(8192);
	assert (p->buffer);

	p->length = floppy_create_directory(floppy,
					    realname, reallength, filetype,
					    secondary, p->buffer);
	p->bufptr = 0;
	return SERIAL_OK;
    }


    /*
     * Now, set filetype according secondary address, if it was not specified
     * on filename
     */

    if (!filetype)
	filetype = (secondary < 2) ? 2 : 1;

    /*
     * Check that there is room on directory.
     */
    set_find_first_slot(floppy, realname, reallength, 0);
    slot = find_next_slot(floppy);

    p->readmode = readmode;


    /*
     * Open file for reading
     */

    if (readmode) {
	int     type;
	int     status;

	if (!slot) {
	    close_1541(floppy, secondary);
	    floppy_error(&floppy->buffers[15], 62, 0, 0);
	    return SERIAL_ERROR;
	}
	type = slot[SLOT_TYPE_OFFSET] & 0x07;

	if (type != filetype) {
	    floppy_error(&floppy->buffers[15], 64, 0, 0);
	    return SERIAL_ERROR;
	}

	/*
	 * Seq, Prg
	 */
	if (type == 1 || type == 2) {
	    p->mode = BUFFER_SEQUENTIAL;
	    p->bufptr = 2;
	    p->buffer = (BYTE *) malloc(256);

	    status = floppy_read_block(floppy->ActiveFd, 
				       p->buffer, (int) slot[SLOT_FIRST_TRACK],
				       (int) slot[SLOT_FIRST_SECTOR]);

	    if (status < 0) {
		close_1541(floppy, secondary);
		return SERIAL_ERROR;
	    }
	    return SERIAL_OK;
	}
	/*
	 * Unsupported filetype
	 */
	return SERIAL_ERROR;
    }


    /*
     * Write
     */

    if (slot) {
	if (*name == '@')
	    remove_slot(floppy, slot);
	else {
	    close_1541(floppy, secondary);
	    floppy_error(&floppy->buffers[15], 63, 0, 0);    /* File exists */
	    return SERIAL_ERROR;
	}
    }

    /*
     * Create slot information. (XXXX: should write directory entry to disk ?)
     */
    p->slot = (BYTE *) malloc(32);
    memset(p->slot, 0, 32);
    memset(p->slot + SLOT_NAME_OFFSET, 0xa0, 16);
    memcpy(p->slot + SLOT_NAME_OFFSET, realname, reallength);
#ifdef DEBUG_1541
    printf("DIR: name (%d) '%s'\n", reallength, realname);
#endif
    p->slot[SLOT_TYPE_OFFSET] = filetype;	/* unclosed */

    p->buffer = (BYTE *) malloc(256);
    p->mode = BUFFER_SEQUENTIAL;
    p->bufptr = 2;

    return SERIAL_OK;
}


int     close_1541(void *flp, int secondary)
{
DRIVE *floppy = (DRIVE *)flp;
    BYTE   *e;
    bufferinfo_t *p = &(floppy->buffers[secondary]);

#ifdef DEBUG_1541
    printf("DEBUG: closing file #%d.\n", secondary);
#endif

    switch (p->mode) {
      case BUFFER_NOT_IN_USE:
	return SERIAL_ERROR;

      case BUFFER_MEMORY_BUFFER:
      case BUFFER_DIRECTORY_READ:
	free((char *)p->buffer);
	p->mode = BUFFER_NOT_IN_USE;
	p->buffer = NULL;
	break;

      case BUFFER_SEQUENTIAL:
	if (!p->readmode) {
	    /*
	     * Flush bytes and write slot to directory
	     */

#ifdef DEBUG_1541
	    printf("DEBUG: flush.\n");
#endif
	    write_sequential_buffer(floppy, p->slot, p->buffer, p->bufptr);

#ifdef DEBUG_1541
	    printf("DEBUG: find empty DIR slot.\n");
#endif
	    set_find_first_slot(floppy, NULL, -1, 0);
	    e = find_next_slot(floppy);

	    if (!e) {
		p->mode = BUFFER_NOT_IN_USE;
		free((char *)p->buffer);
		p->buffer = NULL;

		floppy_error(&floppy->buffers[15], 72, 0, 0);
		return SERIAL_ERROR;
	    }
	    p->slot[SLOT_TYPE_OFFSET] |= 0x80;
	    memcpy(&floppy->find_buffer[floppy->find_slot * 32 + 2],
		p->slot + 2, 30);

#ifdef DEBUG_1541
	    printf("DEBUG: closing, write DIR slot (%d %d) and BAM.\n",
		   floppy->Curr_track, floppy->Curr_sector);
#endif
	    floppy_write_block(floppy->ActiveFd,
			       floppy->find_buffer,
			       floppy->Curr_track,
			       floppy->Curr_sector);

	    floppy_write_block(floppy->ActiveFd, floppy->bam, 18, 0);
	}
	p->mode = BUFFER_NOT_IN_USE;
	free((char *)p->buffer);
	p->buffer = NULL;
	break;

      case BUFFER_COMMAND_CHANNEL:
	floppy_close_all_channels(floppy);
	break;

      default:
	printf("unknown floppy-close-mode\n");
	exit(0);
    }

    return SERIAL_OK;
}


int     read_1541(void *flp, BYTE *data, int secondary)
{
DRIVE *floppy = (DRIVE *)flp;
    bufferinfo_t *p = &(floppy->buffers[secondary]);

#ifdef DEBUG_1541
    if (p->mode == BUFFER_COMMAND_CHANNEL)
	printf("Disk read  %d [%02d %02d]\n", p->mode, 0, 0);
#endif

    switch (p->mode) {
      case BUFFER_NOT_IN_USE:
	floppy_error(&floppy->buffers[15], 61, 0, 0);
	return SERIAL_ERROR;

      case BUFFER_DIRECTORY_READ:
	/*printf("read %d/%d [%02X]\n", p->bufptr, p->length, p->buffer[p->bufptr]);*/
	if (p->bufptr >= p->length)
	    return SERIAL_EOF;
	*data = p->buffer[p->bufptr];
	p->bufptr++;
	break;

      case BUFFER_MEMORY_BUFFER:
	if (p->bufptr >= 256)
	    return SERIAL_EOF;
	*data = p->buffer[p->bufptr];
	p->bufptr++;
	break;

      case BUFFER_SEQUENTIAL:
	if (!p->readmode)
	    return SERIAL_ERROR;

	/*
	 * Read next block if needed
	 */
	if (p->buffer[0]) {
	    if (p->bufptr >= 256) {
		floppy_read_block(floppy->ActiveFd,
				  p->buffer,
				  (int) p->buffer[0],
				  (int) p->buffer[1]);
		p->bufptr = 2;
	    }
	} else {
	    if (p->bufptr >= p->buffer[1])
		return SERIAL_EOF;
	}

	*data = p->buffer[p->bufptr];
	p->bufptr++;
	break;

      case BUFFER_COMMAND_CHANNEL:
	if (p->bufptr > p->length) {
	    floppy_error(&floppy->buffers[15], 0, 0, 0);
	    return SERIAL_EOF;
	}
	*data = p->buffer[p->bufptr];
	p->bufptr++;
	break;

      default:
	printf("unknown buffermode on floppy-read\n");
	exit(0);
    }

    return SERIAL_OK;
}


int     write_1541(void *flp, BYTE data, int secondary)
{
DRIVE *floppy = (DRIVE *)flp;
    int     total = 0;
    bufferinfo_t *p = &(floppy->buffers[secondary]);

#ifdef DEBUG_1541
    if (p -> mode == BUFFER_COMMAND_CHANNEL)
	printf("Disk write %d [%02d %02d] data %02x (%c)\n",
	       p->mode, 0, 0, data, (isprint(data) ? data : '.') );
#endif

    switch (p->mode) {
      case BUFFER_NOT_IN_USE:
	floppy_error(&floppy->buffers[15], 61, 0, 0);
	return SERIAL_ERROR;

      case BUFFER_DIRECTORY_READ:
	floppy_error(&floppy->buffers[15], 60, 0, 0);
	return SERIAL_ERROR;

      case BUFFER_MEMORY_BUFFER:
	if (p->bufptr >= 256)
	    return SERIAL_ERROR;
	p->buffer[p->bufptr] = data;
	p->bufptr++;
	return SERIAL_OK;

      case BUFFER_SEQUENTIAL:
	if (p->readmode)
	    return SERIAL_ERROR;

	if (p->bufptr >= 256) {
	    p->bufptr = 2;
	    if (write_sequential_buffer(floppy,
					p->slot, p->buffer, WRITE_BLOCK) < 0)
		return SERIAL_ERROR;
	}
	p->buffer[p->bufptr] = data;
	p->bufptr++;
	break;

      case BUFFER_COMMAND_CHANNEL:
	if (p->bufptr >= 256)
	    return SERIAL_ERROR;
	p->buffer[p->bufptr] = data;
	p->bufptr++;
	break;

      default:
	printf("Unknown write mode\n");
	exit(-1);
    }

    return total;
}


void    flush_1541(void *flp, int secondary)
{
DRIVE *floppy = (DRIVE *)flp;
    bufferinfo_t *p = &(floppy->buffers[secondary]);

    if (p->mode != BUFFER_COMMAND_CHANNEL)
	return;

    ip_execute(floppy, p->buffer, p->bufptr);
    p->bufptr = 0;
    floppy_error(&floppy->buffers[15], 0, 0, 0);
}


/* ------------------------------------------------------------------------- */

#define IPE_OK		0
#define IPE_DELETED	1
#define IPE_UNIMPL	2

#define IPE_SYNTAX	30
#define IPE_INVAL	31
#define IPE_LONG_LINE	32
#define IPE_BAD_NAME	33
#define IPE_NO_NAME	34

#define IPE_NOT_WRITE	60
#define IPE_NOT_OPEN	61
#define IPE_NOT_FOUND	62
#define IPE_FILE_EXISTS	63
#define IPE_BAD_TYPE	64
#define IPE_NO_BLOCK	65
#define IPE_BAD_TS	66

#define IPE_NO_CHANNEL	70
#define IPE_DISK_FULL	72
#define IPE_DOS_VERSION	73
#define IPE_NOT_READY	74


/*
 * Error messages
 */

typedef struct errortext_s {
    int     nr;
    char   *text;
}       errortext_t;

static errortext_t floppy_error_messages[] =
{
    { 0, "OK"},
    { 1, "FILES SCRATCHED"},
    { 2, "UNIMPLEMENTED"},
    {30, "SYNTAX ERROR"},
    {32, "LINE TOO LONG"},
    {60, "WRITE FILE OPEN"},
    {61, "FILE NOT OPEN"},
    {62, "FILE NOT FOUND"},
    {63, "FILE EXISTS"},
    {64, "FILE TYPE MISMATCH"},
    {70, "NO CHANNEL"},
    {72, "DISK FULL"},
    {73, "TVR 1541 EMULATOR V1.2"},
    {74, "DRIVE NOT READY"},
    {-1, 0}
};


/*
 * Should set values to somewhere so that they could be read from
 * command channel
 */

static void floppy_error(bufferinfo_t *p, int code, int track, int sector)
{
    char   *message;
  /*  bufferinfo_t *p = &(floppy->buffers[15]);*/
    errortext_t *e = &(floppy_error_messages[0]);

    while (e->nr >= 0 && e->nr != code)
	e++;

    if (e->nr >= 0)
	message = e->text;
    else
	message = "UNKNOWN ERROR NUMBER";

    sprintf((char *)p->buffer, "%02d, %s,%d,%d", code, message, track, sector);
    p->length = strlen((char *)p->buffer);
    p->bufptr = 0;

    if (code && code != 73)
	fprintf(stderr, "c1541: ERR = %s\n", message);
}


/* ------------------------------------------------------------------------- */

/*
 * Input Processor Simulator. IP interprets incoming commands and generates
 * the error messages according to return values.
 */


int  ip_execute(DRIVE *floppy, BYTE *buf, int length)
{
    int     status = IPE_SYNTAX;
    BYTE   *p, *p2;
    char   *name;
    BYTE   *minus;
    BYTE   *id;

    if (!length)
	return (IPE_OK);
    if (length > IP_MAX_COMMAND_LEN)
	return (IPE_LONG_LINE);

    p = (BYTE *) malloc(length + 1);
    memcpy(p, buf, length);

    if (p[length-1] == 0x0d) --length;	/* chop CR character */
    p[length] = 0;

    name = (char *)memchr(p, ':', length);
    id   = (BYTE *)memchr(p, ',', length);
    minus= (BYTE *)memchr(p, '-', length);

    if (name)		/* Fix name length */
	for (p2 = p; *p2 && *p2 != ':' && length > 0; p2++, length--);


    switch (p[0]) {
      case 'C':		/* Copy */
	status = do_copy(floppy, (char *)name, length);
	break;

      case 'D':		/* Backup unused */
	status = IPE_SYNTAX;
	break;

      case 'R':		/* Rename */
	break;

      case 'S':		/* Scratch */
	{
	    BYTE *slot;
	    char *realname = name;
	    int   reallength = 0, filetype = 0, readmode = 0;

	    if (floppy_parse_name(name, length, realname, &reallength,
			      &readmode, &filetype) != SERIAL_OK) {
		status = IPE_SYNTAX;
	    }
	    else {
#ifdef DEBUG_1541
		printf("remove name= '%s' len=%d (%d) type= %d.\n",
		       realname, reallength, length, filetype);
#endif
		set_find_first_slot(floppy, realname, reallength, 0);
		slot = find_next_slot(floppy);

		if (slot) {
		    remove_slot(floppy, slot);
		    status = IPE_DELETED;
		}
		else
		    status = IPE_NOT_FOUND;

	    } /* else */
	}
	break;

      case 'I':
	status = do_initialize(floppy);
	break;

      case 'N':
	status = do_format(floppy, name, id, minus);
	break;

      case 'V':
	status = do_validate(floppy);
	break;

      case 'B':		/* Block, Buffer */
	if (!minus || !name)
	    status = IPE_SYNTAX;
	else
	    status = do_block_command(floppy, minus[1], name + 1);
	break;

      case 'M':		/* Memory */
	if (!minus || !name)
	    status = IPE_SYNTAX;
	else
	    /*status = do_memory_command(floppy, minus[1], name + 1);*/
	break;

      case 'P':		/* Position */
	/* 4 byte parameters: channel, rec_lo, rec_hi, pos */
	break;

      case 'U':		/* User */
	switch (p[1]) {
	  case '1':
	  case 'A':
	    if (name)
		status = do_block_command(floppy, 'R', name + 1);
	    break;

	  case '2':
	  case 'B':
	    if (name)
		status = do_block_command(floppy, 'W', name + 1);
	    break;

	  case '3':		/* Jumps */
	  case 'C':
	  case '4':
	  case 'D':
	  case '5':
	  case 'E':
	  case '6':
	  case 'F':
	  case '7':
	  case 'G':
	  case '8':
	  case 'H':
	  case '9':
	    status = IPE_NOT_READY;
	    break;

	  case 'I':		/* Set low speed */
	    status = IPE_OK;
	    break;
	} /* Un */

      default:
	status = IPE_SYNTAX;
	break;
    } /* commands */

    floppy_error(&floppy->buffers[15], status, 0, 0);

    free((char *)p);
    return status;
}


static int  do_block_command(DRIVE *floppy, char command, char *buffer)
{
    int     channel = 0;
    int     drive = 0;
    int     track = 0;
    int     sector = 0;
    int     position = 0;
    int     l;

    switch (command) {
      case 'R':
      case 'W':
	l = sscanf(buffer, "%d %d %d %d", &channel, &drive,
	    &track, &sector);
	if (l < 4) {
	    l = sscanf(buffer, "%d,%d,%d,%d", &channel, &drive,
		&track, &sector);
	    if (l < 4) {
		floppy_error(&floppy->buffers[15], 30, 0, 0);
		return FLOPPY_ERROR;
	    }
	}

#ifdef DEBUG_1541
	printf("B-R/W parsed ok. channel %d mode %d\n",
	       channel, floppy->buffers[channel].mode);
#endif

	if (floppy->buffers[channel].mode != BUFFER_MEMORY_BUFFER) {
	    floppy_error(&floppy->buffers[15], 70, 0, 0);
	    return IPE_NO_CHANNEL;
	}

	if (command == 'W')
	    floppy_write_block(floppy->ActiveFd,
			       floppy->buffers[channel].buffer,
			       track, sector);
	else
	    floppy_read_block(floppy->ActiveFd,
			      floppy->buffers[channel].buffer,
			      track, sector);

	floppy->buffers[channel].bufptr = 0;
	break;

      case 'A':
      case 'F':
	l = sscanf(buffer, "%d %d %d", &drive, &track, &sector);

	if (l < 3) {
	    l = sscanf(buffer, "%d,%d,%d", &drive, &track, &sector);

	    if (l < 3) {
		floppy_error(&floppy->buffers[15], 30, 0, 0);
		return FLOPPY_ERROR;
	    }
	}
	if (command == 'A')
	    allocate_sector(floppy->bam, track, sector);
	else
	    free_sector(floppy->bam, track, sector);
	break;

      case 'P':
	l = sscanf(buffer, "%d %d", &channel, &position);

	if (l < 2) {
	    l = sscanf(buffer, "%d,%d", &channel, &position);

	    if (l < 2) {
		floppy_error(&floppy->buffers[15], 30, 0, 0);
		return FLOPPY_ERROR;
	    }
	}
	floppy->buffers[channel].bufptr = position;
	break;

      default:
	floppy_error(&floppy->buffers[15], 30, 0, 0);
	return FLOPPY_ERROR;
    }
    return FLOPPY_COMMAND_OK;
}


#if 0
static int  do_memory_command(DRIVE *floppy, char command, BYTE *buffer)
{
    int     l;

    int     addrlo = 0;
    int     addrhi = 0;
    int     count = 0;
    int     sector = 0;

    addrlo = buffer[0];
    addrhi = buffer[1];

    switch (command) {
      case 'W':
	count = buffer[2];
	sector= buffer[3];

	if (l < 4) {
	    floppy_error(&floppy->buffers[15], 30, 0, 0);
	    return FLOPPY_ERROR;
	}

	if (floppy->buffers[addrlo].mode != BUFFER_MEMORY_BUFFER) {
	    floppy_error(&floppy->buffers[15], 30, 0, 0);
	    return FLOPPY_ERROR;
	}

	break;

      case 'R':
	if (l < 2) {
	    floppy_error(&floppy->buffers[15], 30, 0, 0);
	    return FLOPPY_ERROR;
	}
	floppy->buffers[channel].bufptr = addrhi;
	break;

      case 'E':
	if (l < 2) {
	    floppy_error(&floppy->buffers[15], 30, 0, 0);
	    return FLOPPY_ERROR;
	}

	floppy->buffers[channel].bufptr = addrhi;
	break;

      default:
	floppy_error(&floppy->buffers[15], 30, 0, 0);
	return FLOPPY_ERROR;
    }
    return FLOPPY_COMMAND_OK;
}
#endif


/* ------------------------------------------------------------------------- */

/*
 * Parse name realname, type and read/write mode. '@' on write must
 * be checked elsewhere
 */

static int  floppy_parse_name(char *name, int length, char *realname, int *reallength, int *readmode, int *filetype)
{
    char   *p;
    char   *c;
    int     t;

    if (!name || !*name)
	return FLOPPY_ERROR;

    p = (char *)memchr(name, ':', length);
    if (p)
	p++;
    else
	p = name;

#ifdef DEBUG_1541
	    printf("Name: '%s'\n", p);
#endif

    if (*name == '@' && p == name)
	p++;

    t = length - (p - name);
    *reallength = 0;
    c = realname;

    while (t > 0 && *p != ',') {
	t--;
	(*reallength)++;
	*c++ = *p++;
    }

    /*
     * Change modes ?
     */
    while (t > 0) {
	t--;
	p++;

	if (t == 0) {
#ifdef DEBUG_1541
	    printf("No type.\n");
#endif
	    return FLOPPY_ERROR;
	}

	switch (*p) {
	  case 'S':
	    *filetype = 1;
	    break;
	  case 'P':
	    *filetype = 2;
	    break;
	  case 'U':
	    *filetype = 3;
	    break;
	  case 'L':	/* L,(#record length) */
	    *filetype = 4;
	    break;
	  case 'R':
	    *readmode = 1;
	    break;
	  case 'W':
	    *readmode = 0;
	    break;
	  default:
#ifdef DEBUG_1541
	    printf("No way. p='%s'\n", p);
#endif
	    return FLOPPY_ERROR;
	}

	c = (char *)memchr(p, ',', t);

	if (c) {
	    t -= (c - p);
	    p = c;
	} else
	    t = 0;
    } /* while t */

    return FLOPPY_COMMAND_OK;
}


/* ------------------------------------------------------------------------- */

/*
 * Create directory listing.
 * If filetype is 0, match for all
 */

static int  floppy_create_directory(DRIVE *floppy, char *name, int length, int filetype, int secondary, BYTE *outputptr)
{
    BYTE   *l, *p;
    BYTE   *origptr = outputptr;
    int     t;
    int     bf;
    int     addr;

    if (*name == '$') { ++name; --length; }
    if (*name == ':') { ++name; --length; }
    if (!*name) { name = "*"; length = 1; }

    l = outputptr;

    /*
     * Start-address
     */
    addr = 0x401;
    SET_LO_HI(l, addr);
    l += 2;

    /*
     * Line number 0
     */
    SET_LO_HI(l, 0);

    /*
     * Reverse on
     */
    *l++ = (BYTE) 0x12;

    *l++ = '\"';
    memcpy(l, &floppy->bam[BAM_DISK_NAME], 16);
    no_a0_pads(l, 16);
    l += 16;
    *l++ = '\"';
    *l++ = ' ';
    memcpy(l, &floppy->bam[BAM_DISK_ID], 5);
    no_a0_pads(l, 5);
    l += 5;

    *l++ = 0;

    /*
     * Address of next line
     */
    addr += ((l - outputptr) - 2);

    outputptr[2] = addr & 0xff;
    outputptr[3] = (addr >>8) & 0xff;

    outputptr = l;


    /*
     * Now, list files that match the pattern.
     */

    set_find_first_slot(floppy, name, length, filetype);

    while ((p = find_next_slot(floppy))) {
	if (p[SLOT_TYPE_OFFSET]) {

	    l += 2;

	    /*
	     * Length and spaces
	     */
	    t = p[SLOT_NR_BLOCKS] + p[SLOT_NR_BLOCKS + 1] * 256;

	    SET_LO_HI(l, t);

	    if (t < 10)
		*l++ = ' ';
	    if (t < 100)
		*l++ = ' ';

	    /*
	     * Filename
	     */
	    *l++ = ' ';
	    *l++ = '\"';

	    memcpy(l, &p[SLOT_NAME_OFFSET], 16);

	    for (t = 0; (t < 16) && (p[SLOT_NAME_OFFSET + t] != 0xa0);)
		t++;

	    no_a0_pads(l, 16);

	    l[16] = ' ';
	    l[t] = '\"';
	    l += 17;

	    /*
	     * Type + End
	     */
	    sprintf ((char *)l, "%c%s%c %c",
		     (p[SLOT_TYPE_OFFSET] & 0x80 ? ' ' : '*'),
		     slot_type[p[SLOT_TYPE_OFFSET] & 0x07],
		     (p[SLOT_TYPE_OFFSET] & 0x40 ? '<' : ' '),
		     0);
	    l += 7;


	    /*
	     * New address
	     */
	    addr += l - outputptr;

	    outputptr[0] = addr & 0xff;
	    outputptr[1] = (addr / 256) & 0xff;

	    outputptr = l;
	}
    }

    /*
     * Calculate blocks free
     */

    bf = 0;
    for (t = 1; t <= 35; t++) {
	if (t != 18)
	    bf += floppy->bam[BAM_BIT_MAP + 4 * (t - 1)];
    }

    *l++ = 0;
    *l++ = 0;
    SET_LO_HI(l, bf);
    memcpy(l, "BLOCKS FREE.", 12);
    l += 12;
    *l++ = (char) 0;

    /* Line Address */
    addr += l - outputptr;
    outputptr[0] = addr & 0xff;
    outputptr[1] = (addr / 256) & 0xff;

    /*
     * end
     */
    *l++ = (char) 0;
    *l++ = (char) 0;
    *l   = (char) 0;

    return (l - origptr);
}


/* ------------------------------------------------------------------------- */


static int  do_copy( DRIVE *floppy, char *dest, int length)
{
    char   *name, *files, *p, c;


    /* Split command line */

    if (!dest || !(files = memchr(dest, '=', length)) )
	return (IPE_SYNTAX);

    *files++ = 0;

    if (strchr (dest, ':'))
	dest = strchr (dest, ':') +1;

#ifdef DEBUG_1541
    printf("COPY: dest= '%s'\n      orig= '%s'\n", dest, files);
#endif

    if (open_1541(floppy, dest, strlen(dest), 1))
	return (IPE_FILE_EXISTS);

    p = name = files;

    while (*name) {				/* loop for given files */
	for (; *p && *p != ','; p++);
	*p++ = 0;

    if (strchr (name, ':'))
	name = strchr (name, ':') +1;

#ifdef DEBUG_1541
	printf("searching for file '%s'\n", name);
#endif
	if (open_1541(floppy, name, strlen(name), 0)) {
	    close_1541(floppy, 1);
	    return (IPE_NOT_FOUND);
	}

	while (!read_1541(floppy, (BYTE *)&c, 0)) {
	    if (write_1541(floppy, c, 1)) {
		close_1541(floppy, 0);		/* no space on disk */
		close_1541(floppy, 1);
		return (IPE_DISK_FULL);
	    }
	}

	close_1541(floppy, 0);
	name = p;				/* next file */
    } /* while */

    close_1541(floppy, 1);
    return(IPE_OK);
}


/* ------------------------------------------------------------------------- */

/*
 * Close all channels. This happens on 'I' -command and on command-
 * channel close
 */

static void floppy_close_all_channels(DRIVE *floppy)
{
    int     i;
    bufferinfo_t *p;

    for (i = 0; i <= 15; i++) {
	p = &(floppy->buffers[i]);

	if (p->mode != BUFFER_NOT_IN_USE && p->mode != BUFFER_COMMAND_CHANNEL)
	    close_1541(floppy, i);
    }
}


static int  do_initialize(DRIVE *floppy)
{
    floppy_close_all_channels(floppy);

    /* Update BAM in memory */
    if (floppy->ActiveFd > 0)
	floppy_read_block(floppy->ActiveFd, floppy->bam, 18, 0);

    return IPE_DOS_VERSION;
}


int  do_validate(DRIVE *floppy)
{
    int     t, s;
    BYTE   *b;
    BYTE    tmp[256];

    do_initialize(floppy);

    b = (BYTE *) floppy->bam + BAM_BIT_MAP;

    for (t = 1; t <= 35; t++) {
	*b++ = 0;
	*b++ = 0;
	*b++ = 0;
	*b++ = 0;

	for (s = 0; s < sector_map[t]; s++)
	    free_sector(floppy->bam, t, s);
    }

    set_find_first_slot(floppy, "*", 1, 0);
/*
	First map out the directory itself.	-Dan Miner
	And another one bites the dust....  :)
*/
	t = 18; s = 0;
	allocate_sector(floppy->bam, t, s);	 /* gec - quick hack to correctly emulate */
	floppy_read_block(floppy->ActiveFd, tmp, t, s);	/* chaining behaviour of the 1541 */
	s++;

	while(t) {
		allocate_sector(floppy->bam, t, s);
		floppy_read_block(floppy->ActiveFd, tmp, t, s);
		t = (int) tmp[0];
		s = (int) tmp[1];
	}

    while ((b = (BYTE *) find_next_slot(floppy))) {
	if (b[SLOT_TYPE_OFFSET]) {
	    floppy->find_buffer[floppy->find_slot * 32 + SLOT_TYPE_OFFSET] |= 0x80;
	    floppy_write_block(floppy->ActiveFd,
			       floppy->find_buffer,
			       floppy->Curr_track,
			       floppy->Curr_sector);
	    t = (int) b[SLOT_FIRST_TRACK];
	    s = (int) b[SLOT_FIRST_SECTOR];

	    while (t) {
		allocate_sector(floppy->bam, t, s);
		floppy_read_block(floppy->ActiveFd, tmp, t, s);
		t = (int) tmp[0];
		s = (int) tmp[1];
	    }
	}
    }

    floppy_write_block(floppy->ActiveFd, floppy->bam, 18, 0);
    return FLOPPY_COMMAND_OK;
}


static int  do_format(DRIVE *floppy, char *name, BYTE *id, BYTE *minus)
{
    BYTE    tmp[256];
    int     status;

    BYTE    null = 0;

    if (!name) {
	floppy_error(&floppy->buffers[15], 30, 0, 0);
	return FLOPPY_ERROR;
    }
    /*
     * If id, skip comma
     */
    if (id)
	*id++ = 0;
    else
	id = &null;

    if (floppy->ErrFlg)
	set_error_data(floppy, 5);	/* clear and write error data */

    /*
     * Make dir-entry
     */
    memset(tmp, 0, 256);
    tmp[1] = 255;
    floppy_write_block(floppy->ActiveFd, tmp, 18, 1);

    /*
     * Make bam
     */
    memset(floppy->bam, 0, 256);
    floppy->bam[0] = 18;
    floppy->bam[1] = 1;
    floppy->bam[2] = 65;
    memset(floppy->bam + BAM_DISK_NAME, 0xa0, 16);
    mystrncpy(floppy->bam + BAM_DISK_NAME, (BYTE *)name + 1, 16);

    memset(floppy->bam + BAM_DISK_ID, 0xa0, 5);
    mystrncpy(floppy->bam + BAM_DISK_ID, id, 2);
    floppy->bam[BAM_VERSION] = 50;
    floppy->bam[BAM_VERSION + 1] = 65;

    status = do_validate(floppy);
    allocate_sector(floppy->bam, 18, 1);	/* first directory block-  Dan Miner */
    floppy_write_block(floppy->ActiveFd, floppy->bam, 18, 0);
    return status;
}


/* ------------------------------------------------------------------------- */
/*
 * Floppy Disc Controller Simulator. FDC executes tasks in the work queue
 * and puts return values there.
 */


/*
 * Check that requested Track and Sector are within limits.
 */

int  check_track_sector(int track, int sector)
{
    int     sectors, i;

    if ((track < 1) || (track > MAX_TRACKS) ||
	(sector < 0) || (sector >= sector_map[track]))
	return (-1);

    for (sectors = 0, i = 1; i < track; i++) {
	sectors += sector_map[i];
    }

    sectors += sector;
    return (sectors);
}


/*
 * Return block offset on 'disk file' according to track and sector
 */

static off_t offset_from_track_and_sector(int track, int sector)
{
    int     sectors;

    if ((sectors = check_track_sector(track, sector)) < 0)
	return -1;

#ifdef DEBUG_1541
    printf("DEBUG SECTOR %3d (%2d,%2d) OFF_T: %ld\n",
	   sectors, track, sector,
	   ((off_t)sectors << 8) + (off_t)HEADER_LENGTH );
#endif

    return ( ((off_t)sectors << 8) + (off_t)HEADER_LENGTH );

}


/*
 * Read one block
 */

static int  floppy_read_block(int fd, BYTE *buf, int track, int sector)
{
    off_t   offset;

    offset = offset_from_track_and_sector(track, sector);

    if (offset < 0)
	return -1;

    lseek(fd, offset, SEEK_SET);

    if (read(fd, (char *)buf, 256) < 256) {
	printf("floppy read error\n");
	return(-2);
    }
    return 0;
}


/*
 * Write one block
 */

static int  floppy_write_block(int fd, BYTE *buf, int track, int sector)
{
    off_t   offset;

    offset = offset_from_track_and_sector(track, sector);

    if (offset < 0)
	return -1;

    lseek(fd, offset, SEEK_SET);

    if (write(fd, (char *)buf, 256) < 256) {
	printf("floppy write error\n");
	return(-2);
    }
    return 0;
}


static int  write_sequential_buffer(DRIVE *floppy, BYTE *slot, BYTE *buf, int length)
{
    static int t = 0, s;
    int     t_new, s_new;
    int     e;

    /*
     * First block of a file ?
     */
    if (t == 0) {
	e = alloc_free_sector(floppy->bam, &t, &s);
	if (e < 0) {
	    floppy_error(&floppy->buffers[15], 72, 0, 0);
	    return -1;
	}
	slot[SLOT_FIRST_TRACK] = t;
	slot[SLOT_FIRST_SECTOR] = s;
    }
    if (length == WRITE_BLOCK) {
	/*
	 * Write current sector and allocate next
	 */
	e = alloc_free_sector(floppy->bam, &t_new, &s_new);
	if (e < 0) {
	    floppy_error(&floppy->buffers[15], 72, 0, 0);
	    return -1;
	}
	buf[0] = t_new;
	buf[1] = s_new;

	floppy_write_block(floppy->ActiveFd, buf, t, s);

	t = t_new;
	s = s_new;

	if (!(++slot[SLOT_NR_BLOCKS]))
	    ++slot[SLOT_NR_BLOCKS + 1];

    } else {
	/*
	 * Write last block
	 */
	buf[0] = 0;
	buf[1] = length - 1;

	floppy_write_block(floppy->ActiveFd, buf, t, s);

	t = 0;

	if (!(++slot[SLOT_NR_BLOCKS]))
	    ++slot[SLOT_NR_BLOCKS + 1];
    }
    return 0;
}


/* ------------------------------------------------------------------------- */

/*
 * Initialize Directory Slot find
 */

static void set_find_first_slot(DRIVE *floppy, char *name, int length, int type)
{
    floppy->find_name   = name;
    floppy->find_type   = type;
    floppy->find_length = length;

#ifdef DEBUG_1541
    printf("Prepare seeking entry for '%s' of type %d. (len =%d)\n",
	   (name ? name : "<noname>"),
	   type, length);
#endif

    /*
     * Make sure find_next_slot() loads first directory block ...
     * Emulate chaining behaviour of the 1541  -- GEC
     */

    floppy->Curr_track  = 18;
    floppy->Curr_sector =  1;

    floppy->find_slot = -1;
    floppy_read_block(floppy->ActiveFd, floppy->find_buffer, 18, 1);
}


static BYTE *find_next_slot(DRIVE *floppy)
{
    static BYTE return_slot[32];
    int     status;

    floppy->find_slot++;

#ifdef DEBUG_1541
    printf("find next slot for '%s' of type %d. slot: %d\n",
	   (floppy->find_name ? floppy->find_name : "<noname>"),
	   floppy->find_type, floppy->find_slot);
#endif

    /*
     * Loop all directory blocks starting from track 18, sector 1
     */

    do {
	/*
	 * Load next(first) directory block ?
	 */

	if (floppy->find_slot >= 8) {
	    if (!(*(floppy->find_buffer))) {
#ifdef DEBUG_1541
		printf("error.\n");
#endif
		return NULL;
	    }

	    floppy->find_slot = 0;

	    floppy->Curr_track  = (int) floppy->find_buffer[0];
	    floppy->Curr_sector = (int) floppy->find_buffer[1];

	    status = floppy_read_block(floppy->ActiveFd,
				       floppy->find_buffer,
				       floppy->Curr_track,
				       floppy->Curr_sector);
	}

	while (floppy->find_slot < 8) {
	    if (floppy_name_match(&floppy->find_buffer[floppy->find_slot * 32],
		    floppy->find_name, floppy->find_length,
		    floppy->find_type)) {
		memcpy(return_slot,
		    &floppy->find_buffer[floppy->find_slot * 32], 32);
#ifdef DEBUG_1541
		printf("found slot %d  on %d %d.\n",
		       floppy->find_slot,
		       floppy->Curr_track,
		       floppy->Curr_sector);
#endif
		return return_slot;
	    }
	    floppy->find_slot++;
	}
    } while (*(floppy->find_buffer));


    /*
     * If length < 0, create new directory-entry if possible
     */

    if (floppy->find_length < 0) {
	int     s;

#ifdef DEBUG_1541
		printf("create a new entry.\n");
#endif

	for (s = 1; s < sector_map[18]; s++) {
	    if (allocate_sector(floppy->bam, 18, s)) {
		floppy->find_buffer[0] = 18;
		floppy->find_buffer[1] = s;
		floppy_write_block(floppy->ActiveFd,
				   floppy->find_buffer,
				   floppy->Curr_track,
				   floppy->Curr_sector);

#ifdef DEBUG_1541
		printf("Found (%d %d) TR = %d SE = %d.\n",
		       18, s,
		       floppy->Curr_track,
		       floppy->Curr_sector);
#endif

		floppy->find_slot = 0;
		memset(floppy->find_buffer, 0, 256);
		floppy->find_buffer[1] = 0xff;
		floppy->Curr_sector = s;
		return floppy->find_buffer;
	    }
	}

    } /* length */

    return NULL;
}


/*
 * If type is 0, check next slot with same name for match.
 * If type is nonzero, match next that has the same id and name.
 */

static int  floppy_name_match(BYTE *slot, char *name, int length, int type)
{
    int     i;

    if (length < 0) {
	if (slot[SLOT_TYPE_OFFSET])
	    return 0;
	else
	    return 1;
    }
    if (!slot[SLOT_TYPE_OFFSET])
	return 0;

    if (length > 16)
	length = 16;

    for (i = 0; i < length; i++) {
	switch (name[i]) {
	  case '?':
	    /*
	     * Match any character
	     */
	    break;

	  case '*':
	    /*
	     * Make a match
	     */
	    i = 16;
	    break;

	  default:
	    if ((BYTE)name[i] != slot[i + SLOT_NAME_OFFSET])
		return 0;
	}
    }

    /*
     * Got name match ?
     */
    if (i < 16 && slot[SLOT_NAME_OFFSET + i] != 0xa0)
	return 0;

    if (type && type != (slot[SLOT_TYPE_OFFSET] & 0x07))
	return 0;

    return 1;
}


/*
 * remove_slot() is called from open_1541() (in 'save and replace')
 * and from ip_execute() for 'SCRATCH'.
 */

static void remove_slot(DRIVE *floppy, BYTE *slot)
{
    BYTE    buf[256];
    int     tmp;
    int     t, s;

    /*
     * Find slot
     */
    for (tmp = 0; (tmp < 16) && slot[SLOT_NAME_OFFSET + tmp] != 0xa0; tmp++);

    set_find_first_slot(floppy,
			(char *)&slot[SLOT_NAME_OFFSET], tmp,
			slot[SLOT_TYPE_OFFSET] & 0x07);

    /*
     * If slot slot found, remove
     */
    if (find_next_slot(floppy)) {

	/*
	 * Free all buffers from bam
	 */
	t = (int) floppy->find_buffer[floppy->find_slot * 32 + SLOT_FIRST_TRACK];
	s = (int) floppy->find_buffer[floppy->find_slot * 32 + SLOT_FIRST_SECTOR];

	while (t) {
	    free_sector(floppy->bam, t, s);
	    floppy_read_block(floppy->ActiveFd, buf, t, s);
	    t = (int) buf[0];
	    s = (int) buf[1];
	}

	/* Update bam */
	floppy_write_block(floppy->ActiveFd, floppy->bam, 18, 0);

	/* Update directory entry */
	floppy->find_buffer[floppy->find_slot * 32 + SLOT_TYPE_OFFSET] = 0;
	floppy_write_block(floppy->ActiveFd, 
			   floppy->find_buffer,
			   floppy->Curr_track,
			   floppy->Curr_sector);
    }
}


/* ------------------------------------------------------------------------- */

static int  alloc_free_sector(BYTE *bam, int *track, int *sector)
{
    int     t, s;

    for (t = 17; t >= 1; t--) {
	for (s = 0; s < sector_map[t]; s++) {
	    if (allocate_sector(bam, t, s)) {
		*track = t;
		*sector = s;
		return 0;
	    }
	}
    }

    for (t = 19; t <= 35; t++) {
	for (s = 0; s < sector_map[t]; s++) {
	    if (allocate_sector(bam, t, s)) {
		*track = t;
		*sector = s;
		return 0;
	    }
	}
    }
    return -1;
}


static int  allocate_sector(BYTE *bam, int track, int sector)
{
    BYTE   *bamp;	/* Macros use this */

    bamp = &bam[BAM_BIT_MAP + 4 * (track - 1)];

    if (BAM_ISSET(sector)) {
	(*bamp)--;
	BAM_CLR(sector);

	return 1;
    }
    return 0;
}


static int  free_sector(BYTE *bam, int track, int sector)
{
    BYTE   *bamp;

    bamp = &bam[BAM_BIT_MAP + 4 * (track - 1)];

    if (!(BAM_ISSET(sector))) {
	BAM_SET(sector);
	(*bamp)++;
	return 1;
    }
    return 0;
}


/* ------------------------------------------------------------------------- */

/*
 * Functions to attach the disk image files.
 */


void    detach_floppy_image(DRIVE *floppy)
{

    floppy_close_all_channels(floppy);

    if (floppy->ActiveFd >= 0) {
	close(floppy->ActiveFd);
	floppy->ActiveFd = -1;
	printf("Detaching floppyimage %s\n", floppy->ActiveName);
    }
}


int     attach_floppy_image(DRIVE *floppy, char *name, int mode)
{
    int     fd;
    hdrinfo hdr;

    if ( !(floppy -> type & DT_DISK)) {
	printf("Incompatible emulator mode: FS 1541 Drive.\n");
	return (-1);
    }

    detach_floppy_image(floppy);

    if (!name) {
	printf("No name, detaching floppyimage\n");
	return (-1);
    }

#ifdef C64_MSDOS
    fd = open(name, O_RDWR | O_BINARY);
#else
    fd = open(name, O_RDWR);
#endif

    if (fd >= 0) {
	if (check_header(fd, &hdr)) {
	    printf("file '%s' does not have valid C1541 header\n", name);
	    close(fd);
	    return (-2);
	}
	if (hdr.v_major > C1541_VERSION_MAJOR ||
	    (hdr.v_major == C1541_VERSION_MAJOR && hdr.v_minor > C1541_VERSION_MINOR)) {
	    printf("c1541 image file %s (V %d.%02d) newer that emulator (V %d.%02d)\n",
		name, hdr.v_major, hdr.v_minor, C1541_VERSION_MAJOR, C1541_VERSION_MINOR);
	    close(fd);
	    return (-2);
	}

	if (hdr.devtype != (floppy-> type & DT_MASK))
	    fprintf(stderr, "Warning: disk drive type mismatch\n");

	/* mode defines what to do if disk image is altered or incompatible */


	floppy->ActiveFd = fd;
	strcpy(floppy->ActiveName, name);
	floppy->changed  = 0;

	floppy->NumTracks = hdr.tracks;
	floppy->NumBlocks = ((hdr.tracks == 35) ? 683 : 768);
	floppy->ErrFlg    = hdr.errblk;

	if (hdr.errblk) {
	    set_error_data(floppy, 3);	/* clear or read error data */
	}

	floppy_read_block(floppy->ActiveFd, floppy->bam, 18, 0);

	printf("c1541 image file %s (version %d.%02d) attached\n", name,
	    hdr.v_major, hdr.v_minor);
    }

    return (0);
}


int  check_header(int fd, hdrinfo *hdr)
{
    BYTE    header[HEADER_LENGTH];

    lseek (fd, (off_t) 0, SEEK_SET);
    if (read(fd, (BYTE *)header, sizeof (header)) != sizeof (header)) {
	printf("cannot read image header\n");
	return FD_RDERR;
    }

    if (header[HEADER_MAGIC_OFFSET + 0] != HEADER_MAGIC_1 ||
	header[HEADER_MAGIC_OFFSET + 1] != HEADER_MAGIC_2 ||
	header[HEADER_MAGIC_OFFSET + 2] != HEADER_MAGIC_3 ||
	header[HEADER_MAGIC_OFFSET + 3] != HEADER_MAGIC_4) {

	return FD_BADIMAGE;
    }

    hdr-> v_major = header[HEADER_VERSION_OFFSET + 0];
    hdr-> v_minor = header[HEADER_VERSION_OFFSET + 1];

  /* Disk type flags: Device Type, Max Tracks, Side, and Error Flag. */
    hdr-> devtype = header[HEADER_FLAGS_OFFSET];
    hdr-> tracks  = header[HEADER_FLAGS_OFFSET + 1];
    /*hdr-> sides = header[HEADER_FLAGS_OFFSET + 2];*/
    hdr-> errblk  = header[HEADER_FLAGS_OFFSET + 3];

    if (hdr-> tracks == 0)
	hdr-> tracks = 35;

    strncpy (hdr->description,
	     (char *)header + HEADER_LABEL_OFFSET, HEADER_LABEL_LEN);

    return FD_OK;
}


/*
 * Sync (update) the Error Block image in memory
 */

static int  set_error_data(DRIVE *floppy, int flags)
{
    int    size;
    off_t  offset;


    if (!(floppy->ErrData)) {
	floppy->ErrData = (char *)malloc((size_t)768);
	assert (floppy->ErrData);
	memset(floppy->ErrData, 0x01, 768);
    }
    else if (flags & 1)		/* clear error data */
	memset(floppy->ErrData, 0x01, 768);


    size   = ((floppy->NumTracks == 35) ? 683 : 768);
    offset = ((off_t)size << 8) + (off_t)HEADER_LENGTH;
    lseek(floppy->ActiveFd, offset, SEEK_SET);

    if (flags & 2) {		/* read from file */
	if (read(floppy->ActiveFd, (char *)floppy->ErrData, size) < size) {
	    printf("floppy read error\n");
	    return(-2);
	}
    }

    if (flags & 4) {		/* write to file */
	if (write(floppy->ActiveFd, (char *)floppy->ErrData, size) < size) {
	    printf("floppy write error\n");
	    return(-2);
	}
    }

    return (0);
}


/*
 * Return 'error' according to track and sector
 */

static int  get_disk_error(DRIVE *floppy, int track, int sector)
{
    int     sectors, i;

    if ((sectors = check_track_sector(track, sector)) < 0)
	return -1;

    if (floppy->ErrFlg) {
	i = floppy->ErrData[sectors];
    }
    else
	i = 0x01;


    return (i);
}


/*
 * Misc functions
 */

static int  mystrncpy(BYTE *d, BYTE *s, int n)
{
    while (n-- && *s)
	*d++ = *s++;
    return (n);
}


static void no_a0_pads(BYTE *ptr, int l)
{
    while (l--) {
	if (*ptr == 0xa0)
	    *ptr = 0x20;

	ptr++;
    }
}


/* ------------------------------------------------------------------------- */

/*
 * File system interface
 */

enum fsmode {Write, Read, Directory };

struct fs_buffer_info {
    FILE   *fd;
    DIR    *dp;
    enum fsmode mode;
    char    dir[MAXPATHLEN];
    BYTE    buf[MAXPATHLEN + 5];
    int     buflen;
    BYTE   *bufp;
    int     eof;
}       fs_info[16];


int     write_fs(void *flp, BYTE data, int secondary)
{
    if (fs_info[secondary].mode != Write)
	return FLOPPY_ERROR;

    if (fs_info[secondary].fd) {
	fputc(data, fs_info[secondary].fd);
	return FLOPPY_COMMAND_OK;
    };

    return FLOPPY_ERROR;
}


int     read_fs(void *flp, BYTE *data, int secondary)
{
    int     i;
    unsigned short blocks;
    struct dirent *dirp;
    struct stat statbuf;
    struct fs_buffer_info *info = &fs_info[secondary];

    switch (info->mode) {
      case Write:
	return FLOPPY_ERROR;

      case Read:
	if (info->fd) {
	    i = fgetc(info->fd);
	    if (ferror(info->fd))
		return FLOPPY_ERROR;
	    if (feof(info->fd))
		return SERIAL_EOF;
	    *data = i;
	    return SERIAL_OK;
	}
	break;

      case Directory:
	if (info->dp) {
	    if (info->buflen <= 0) {
		char    buf[MAXPATHLEN];
		if (info->eof)
		    return SERIAL_EOF;

		if ((dirp = readdir(info->dp))) {
		    BYTE  *p = info->buf;
		    int     j;

		    strcpy(buf, info->dir);
		    strcat(buf, "/");
		    strcat(buf, dirp->d_name);
		    if (stat(buf, &statbuf) < 0)
			return FLOPPY_ERROR;

		    *p++ = 1;
		    *p++ = 4;
		    blocks = (unsigned short) (statbuf.st_size / 256);
		    *p++ = LOWER(blocks);
		    *p++ = UPPER(blocks);
		    for (j = 0; blocks > 9; j++)
			blocks /= 10;
		    i = 4 - j;
		    while (i-- > 0)
			*p++ = ' ';
		    *p++ = '"';

		    for (i = 0; i < dirp->d_namlen; i++)
			*p++ = p_topetcii(dirp->d_name[i]);

		    *p++ = '"';
		    *p++ = '\0';

		    info->buflen = (int) p - (int) info->buf;

		} else {
		    info->buf[0] = 0;
		    info->buf[1] = 0;
		    info->buf[2] = 0;
		    info->buflen = 3;
		    info->eof++;
		}
		info->bufp = info->buf;
	    }
	    *data = *info->bufp++;
	    info->buflen--;
	    return SERIAL_OK;
	}
	break;
    }

    return FLOPPY_ERROR;
}


int     open_fs(void *flp, char *name, int length, int secondary)
{
    FILE   *fd;
    DIR    *dp;
    char    tmp[256], *p;

    if (fs_info[secondary].fd)
	return FLOPPY_ERROR;

    /*
     * Support only load / save
     */
    if (secondary < 0 || secondary >= 2)
	return FLOPPY_ERROR;

    memcpy(tmp, name, length);
    tmp[length] = 0;
    petconvstring(tmp, 0);
    
    if (*tmp == '$') {
	if (secondary != 0)
	    return FLOPPY_ERROR;

	if (tmp[1]) {
	    if (!(dp = opendir((char *)tmp + 1))) {
		for (p = tmp; *p; p++)
		    if (isupper(*p))
			*p = tolower(*p);
		if (!(dp = opendir((char *)tmp + 1)))
		    return FLOPPY_ERROR;
	    }
	    strcpy(fs_info[secondary].dir, tmp + 1);
	} else {
	    if (!(dp = opendir(".")))
		return FLOPPY_ERROR;
	    strcpy(fs_info[secondary].dir, ".");
	}

	fs_info[secondary].mode = Directory;
	fs_info[secondary].dp = dp;
	fs_info[secondary].eof = 0;
	fs_info[secondary].buflen = 2;
	fs_info[secondary].bufp = fs_info[secondary].buf;
	fs_info[secondary].buf[0] = 0;
	fs_info[secondary].buf[1] = 4;

    } else {
	fs_info[secondary].mode = (secondary == 1 ? Write : Read);
	fd = fopen(tmp, fs_info[secondary].mode ? READ : APPEND);

	if (!fd) {
	    for (p = tmp; *p; p++)
		if (isupper(*p))
		    *p = tolower(*p);

	    fd = fopen(tmp, fs_info[secondary].mode ? READ : APPEND);

	    if (!fd)
		return FLOPPY_ERROR;
	}
	fs_info[secondary].fd = fd;
    }
    return FLOPPY_COMMAND_OK;
}


int     close_fs(void *flp, int secondary)
{
    switch (fs_info[secondary].mode) {
      case Write:
      case Read:
	if (!fs_info[secondary].fd)
	    return FLOPPY_ERROR;

	fclose(fs_info[secondary].fd);
	fs_info[secondary].fd = NULL;
	break;

      case Directory:
	if (!fs_info[secondary].dp)
	    return FLOPPY_ERROR;

	closedir(fs_info[secondary].dp);
	fs_info[secondary].dp = NULL;
	break;
    }

    return FLOPPY_COMMAND_OK;
}
