/*
 * $Id: interrupt.c,v 1.3 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains interrupt handling
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta     (jopi@stekt.oulu.fi)
 *   Hannu Helminen  (dm@stekt.oulu.fi)
 *
 *
 * $Log: interrupt.c,v $
 * Revision 1.3  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Bug in interrupt vs. RTI fixed.
 *
 * Revision 1.2  1994/12/12  16:59:44  jopi
 * 16-bit portability
 *
 * Revision 1.1  1994/06/16  17:19:26  jopi
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <time.h>

#include "macro.h"
#include "vmachine.h"
#include "extern.h"


/* global */

int     next_int;
CLOCK   next_int_clk;
CLOCK   int_clk[NUMOFINT];



/* find next interrupt */

void    find_next_int(void)
{
    int     i;

    for (i = 0, next_int_clk = (CLOCK) - 1; i < NUMOFINT; i++)
	if (int_clk[i] && int_clk[i] < next_int_clk) {
	    next_int_clk = int_clk[i];
	    next_int = i;
	}
    if (debugflg && next_int != I_SPECX)
	printf("%ld next int %d %ld\n", clk, next_int, next_int_clk);
}


void    restore(void)
{
    int_clk[I_RESTORE] = clk;
    find_next_int();
#ifdef NEW_TIMER_CODE
    update_intr_alarm();
#endif
}


/*
 * Serve interrupts. Called from CPU mainloop.
 */

void test_irq(CLOCK ck)
{
#ifdef NEW_TIMER_CODE
    if (ck >= next_alarm_clk)
	alarm_handler();
#else
    while (ck > next_int_clk)
	PC = intr(next_int, PC);
#endif
}


/* interrupts */

ADDRESS intr(int sig, ADDRESS pc)
{
    BYTE    t;

    switch (sig) {
      case I_SPECX:
	/* special "interrupt" to handle x events */
	break;

      case I_RESET:
	ram[0] = 0xff;
	ram[1] = 0xff;
	screen_init();
	store_vic(0x18, 0x15);
	cia2[2] = 0;
	cia2[0] = 255;
	clk = 0;
	init_memmaps();		/* RESET */
	if (debugflg)
	    printf("RESET\n");
	pc = LOAD_ADDR(0xFFFC);	/* Jump using RESET vector. */
	break;

      case I_BRK:
	SET_BREAK(1);		/* Set break status. */
	PUSH(UPPER(pc));
	PUSH(LOWER(pc));	/* Push return address into the stack. */
	t = GET_SR();
	PUSH(t);		/* Push status register into the stack. */
	SET_INTERRUPT(1);	/* Disable interrupts. */
	pc = LOAD_ADDR((ADDRESS)65534);	/* Jump using BRK vector (0xfffe). */
	return (pc);

      case I_CIA1TA:
      case I_CIA1TB:
/*printf("TIMER IRQ\n");*/
#ifndef C128
	x_get_keyarr();
      case I_RASTER:
#else
      case I_RASTER: /* C128 uses raster instead */
	/*printf("IRQ %ld\n", clk);*/
	x_get_keyarr();
#endif
      case I_LIGHT_PEN:
      case I_SS_COL:
      case I_SB_COL:

      case I_CIA1FL:
      case I_CIA1SP:
      case I_CIA1TOD:		/* check if IRQ iterrupts are not masked */
	if (IF_INTERRUPT())
	    break;

	SET_BREAK(0);
	PUSH(UPPER(pc));
	PUSH(LOWER(pc));	/* Push return address into the stack. */
	t = GET_SR();
	PUSH(t);		/* Push status register into the stack. */
	SET_INTERRUPT(1);	/* Disable interrupts. */
	pc = LOAD_ADDR((ADDRESS)65534);	/* Jump using IRQ vector (0xfffe). */
	break;

      case I_CIA2FL:
      case I_CIA2SP:
      case I_CIA2TOD:
      case I_CIA2TA:
      case I_CIA2TB:
      case I_RESTORE:
	SET_BREAK(0);
	PUSH(UPPER(pc));
	PUSH(LOWER(pc));	/* Push return address into the stack. */
	t = GET_SR();
	PUSH(t);		/* Push status register into the stack. */
	SET_INTERRUPT(1);	/* Disable interrupts. */
	pc = LOAD_ADDR(0xFFFA);	/* Jump using NMI vector. */
	break;
    }

    /*
     * we must handle interrupts allthough they are not served eg. interrupts
     * are disabled, so that timers are updated
     */

    int_clk[sig] = 0;

    switch (sig) {
      case I_SPECX:
	int_clk[sig] = clk + 4 * CYCLES_PER_RFSH;
	x_loop();
#ifdef JOYSTICK
	joystick();
#endif
	pc = PC;	/* PC may be changed in x_loop() or mon() */
	break;

      case I_RASTER:
	int_raster();
	break;

      case I_CIA1TB:
	printf("cia1tbint\n");
	int_cia1tb();
	break;

      case I_CIA1TA:
	int_cia1ta();
	break;

      case I_CIA2TB:
	printf("cia2tbint\n");
	int_cia2tb();
	break;

      case I_CIA2TA:
	printf("cia2taint\n");
	/* int_cia2ta(); */
	break;

      case I_LIGHT_PEN:
      case I_SS_COL:
      case I_SB_COL:

      case I_CIA1FL:
      case I_CIA1SP:

      case I_CIA2FL:
      case I_CIA2SP:

	printf("unsupported interrupt %d\n", sig);
	show();
	end64();
    }

    find_next_int();
    return (pc);
}


/*
 * This routine is used to skip idle on ROM busy loop
 */
void    jump_to_next_alarm()
{
    clk = next_int_clk;
}


#ifdef NEW_TIMER_CODE
/*
 * This is part kludge to get both timers usable
 */
/* static */ int new_alarm_number;


void new_alarm_trap(int i)
{
    while (clk > next_int_clk)
	PC = intr(next_int, PC);
    set_alarm(new_alarm_number, next_int_clk+1, new_alarm_trap, 0);
}


void update_intr_alarm(void)
{
    set_alarm(new_alarm_number, next_int_clk+1, new_alarm_trap, 0);
    find_next_alarm();
}

#endif /* NEW_TIMER_CODE */
