/*
 * $Id: memory.c,v 1.10 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains cpu <> memory <> io interface. Memory
 * is divided into blocks which are of $1000 (4k) lenght each.
 * memwr and memrd are tables for memory access functions and
 * screen drawing is done by putting hook into a block.
 *
 * Other things:
 *	o Also memory banking is handled here.
 *	o Resource files allocating, loading and storing sits here as well.
 *	o ROM checksums
 *	o Emulator identification
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Teemu Rantanen  (tvr@cs.hut.fi)
 *   Jouko Valta     (jopi@stekt.oulu.fi)
 *
 * Patch by
 *   Hannu Helminen  (dm@stekt.oulu.fi)
 *
 *
 * $Log: memory.c,v $
 * Revision 1.10  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * C128 ROM checksums done better way. More error messages.
 * Support for VDC emulation. Basic relinking fixed.
 *
 * Revision 1.9  1994/12/12  16:59:44  jopi
 * 16-bit portability
 *
 * Revision 1.8  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.7  1994/06/07  18:31:11  jopi
 * Patchlevel 2
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 * New debugger & arg parser, x20, and x128
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 * x_get_keyarr moved to IRQ
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:39  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>  /* only for convenience */

#include "macro.h"
#include "extern.h"
#include "vmachine.h"
#include "memory.h"


/* global */

BYTE    basic_rom[LOROM_SIZE];
BYTE    kernal_rom[HIROM_SIZE];
BYTE   *external_rom = 0;
BYTE    ram[RAMSIZE];
BYTE    sid[32];

BYTE   *allocate_mem( int size );
char   *create_name( char *path, char *name );
int     Load_Sys_File( char *path, char *name, BYTE *var, int reloc, int siz );
int     LoadFile( char *path, char *name, BYTE *var, int reloc, int maxsiz );
int     SaveFile( char *path, char *name, BYTE *var, int reloc, int siz );
void    set_text_pointers( ADDRESS ls, ADDRESS end );

extern unsigned char read_vic ( ADDRESS addr );
extern void store_vic( ADDRESS addr, BYTE byte );
extern void store_colorram( ADDRESS addr, BYTE byte );
extern BYTE read_colorram( ADDRESS addr );


#ifdef C128
BYTE   mmu[11];
static BYTE   vdc_ptr = 0;
#ifdef USE_VDC
extern void store_vdc ( BYTE reg, BYTE byte );
extern BYTE read_vdc ( BYTE reg );
#else
static BYTE   vdc[37];		/* Needed to trick the system */
#endif

#if (RAMSIZE == 0x40000)
#define RAMBANK()  (((long)mmu[0] & 0xC0) << 10)  /* 256K MMU support */
#else
#define RAMBANK()  (((long)mmu[0] & 0x40) << 10)  /* 128K only */
#endif
#define RAMSHAR(a)  ((a < 0x400) ? a : RAMBANK() + a)

#else  /* C128 */
#define RAMSHAR(a)  (a)

#endif /* C128 */

#if 0
#define dprintf(a)  printf(a)
#else
#define dprintf(a)
#endif

#ifdef SYSV
#define mAdj 0x92
#else
#ifdef DEC
#define mAdj 0x2df2
#else
#define mAdj 0x1a
#endif
#endif

/* function tables to access memory */

void    (*(memwr[17])) ( ADDRESS, BYTE );
BYTE    (*(memrd[17])) ( ADDRESS );

extern  void iowr ( ADDRESS, BYTE );
extern  BYTE iord ( ADDRESS );

/* local functions */
static void ramwr ( ADDRESS, BYTE );
static void zerowr ( ADDRESS, BYTE );

#if (defined(REU) || defined(C128))
static void ramwr_hi ( ADDRESS, BYTE );
#endif
#ifdef C128
static void store_mmu ( ADDRESS, BYTE );
#endif

static BYTE ramrd ( ADDRESS );
static BYTE basicrd ( ADDRESS );
static BYTE kernelrd ( ADDRESS );
static BYTE exromrd ( ADDRESS );
static BYTE charrd ( ADDRESS );


#ifdef PATCH_ROM
static BYTE inittab1[] = {
    47, 111, 119, 108, 115, 47, 97, 114, 101, 47,
    110, 111, 116, 47, 119, 104, 97, 116, 47, 116,
    104, 101, 121, 47, 115, 101, 101, 109, 47, 10
};
#endif

#ifdef C128
static BYTE inittab0[] = {
    0x4e, 0x67, 0x7c, 0x64, 0x6e, 0x7a, 0x62, 0x7e, 0x19, 0x2f,
    0x01, 0x4a, 0x09, 0x06, 0x53, 0x65, 0x57, 0x4e, 0x58, 0x15,
    0x3f, 0x74, 0x33, 0x3c, 0x60, 0x5d, 0x5f, 0x5e, 0x5e, 0x58,
    0x50, 0x5a, 0x06, 0x2a,
    0x7d, 0x26, 0x2f,  0x7c, 0x4c, 0x42, 0x57, 0x43, 0x4f, 0x45, 0x49,
    0x0a
};
#endif

#ifdef EMULATOR_ID
static BYTE Emulator_ID[] = {
   0x58, 0x36, 0x34, 0x20, 0x30, 0x2e, 0x33, 0x0d,
   0x28, 0x43, 0x29, 0x31, 0x39, 0x39, 0x33, 0x2d,	/* X64 0.3.(C)1993- */
   0x39, 0x35, 0x20, 0x4a, 0x2e, 0x53, 0x4f, 0x4e,
   0x4e, 0x49, 0x4e, 0x45, 0x4e, 0x2f, 0x54, 0x2e,	/* 95 J.SONNINEN/T. */
   0x52, 0x41, 0x4e, 0x54, 0x41, 0x4e, 0x45, 0x4e,
   0x2f, 0x4a, 0x2e, 0x56, 0x41, 0x4c, 0x54, 0x41,	/* RANTANEN/J.VALTA */
   0x20, 0x75, 0x4e, 0x44, 0x45, 0x52, 0x20, 0x67,
   0x6e, 0x75, 0x20, 0x67, 0x45, 0x4e, 0x45, 0x52,	/*  uNDER gnu gENER */
   0x41, 0x4c, 0x20, 0x70, 0x55, 0x42, 0x4c, 0x49,
   0x43, 0x20, 0x6c, 0x49, 0x43, 0x45, 0x4e, 0x53,	/* AL pUBLIC lICENS */
   0x45, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x58, 0x55	/* E............0XU */
};
#endif


/* ------------------------------------------------------------------------- */

/*
 * The following functions are used for installing
 * traps and patches to the operating system.
 */

BYTE    load_rom(ADDRESS a)
{
    switch(a & 0xF000)
    {
#ifdef C128
    case 0x4000:
    case 0x5000:
    case 0x6000:
    case 0x7000:
    case 0x8000:
    case 0x9000:
    case 0xA000:
    case 0xB000:
    case 0xC000:	/* C000 is in the C128 Kernel ROM */
#else
#ifdef VIC20
    case 0xC000:
    case 0xD000:
#else  /* C64 */
    case 0xA000:
    case 0xB000:
#endif
#endif
	return basic_rom[BASIC_ADDRESS(a)];

#ifdef VIC20
    case 0x8000:
#else
    case 0xD000:
#endif
	return char_rom[a & 0x0FFF];

    case 0xE000:
    case 0xF000:
	return kernal_rom[a & 0x1FFF];
    }
    return 0;
}


void    save_rom(ADDRESS a, BYTE x)
{
    switch(a & 0xF000)
    {
#ifdef C128
    case 0x4000:
    case 0x5000:
    case 0x6000:
    case 0x7000:
    case 0x8000:
    case 0x9000:
    case 0xA000:
    case 0xB000:
    case 0xC000:
#else
#ifdef VIC20
    case 0xC000:
    case 0xD000:
#else  /* C64 */
    case 0xA000:
    case 0xB000:
#endif
#endif
	basic_rom[BASIC_ADDRESS(a)] = x;
	break;

#ifdef VIC20
    case 0x8000:
#else
    case 0xD000:
#endif
	char_rom[a & 0x0FFF] = x;
	break;

    case 0xE000:
    case 0xF000:
	kernal_rom[a & 0x1FFF] = x;
	break;
    }
}


/* ------------------------------------------------------------------------- */

/* Load and test memory image files (at boot up). */

int     load_mem(void)
{
    int   i;
    unsigned short s;
    BYTE  id;

 /*printf("%s\n%s\n%s\n%s\n",
	app_resources.directory, app_resources.ramName,
	app_resources.kernelName, app_resources.basicName);*/

    if (Load_Sys_File (NULL, app_resources.ramName, ram, 0, RAMSIZE) < 0) {
        fprintf(stderr,"Warning, RAM image not loaded\n");
	autodump++;

	if (module) {
	    fprintf (stderr, "Cannot use module without ram image. Please dump ram.\n");
	    module = 0;
	}
    }

    if (Load_Sys_File (NULL, app_resources.kernelName, kernal_rom,
		       0xe000, HIROM_SIZE) < 0) {
	fprintf(stderr,"Couldn't load kernel.\n\n");
	fprintf(stderr,"See Section 7 in README for details.\n\n");
	return -1;
    }

    /* Check kernel
     * There are several Kernel versions for different purposes.
     */

    for (i = 0, s = 0; i < HIROM_SIZE; i++) s += kernal_rom[i];

#ifdef VIC20
    if (s != KERNEL_CHECKSUM)
#else
    id = kernal_rom[0x1f80];
    printf("Kernel rev #%d\n", id);

#ifdef C128
    if (id == 01 &&
	s != KERNEL_CHECKSUM_R01 &&
	s != KERNEL_CHECKSUM_R01swe && s != KERNEL_CHECKSUM_R01ger)
#else
    if ((id == 0 && s != KERNEL_CHECKSUM_R00) ||
	(id == 03 &&
	  s != KERNEL_CHECKSUM_R03 && s != KERNEL_CHECKSUM_R03swe) ||
	(id == 0x43 && s != KERNEL_CHECKSUM_R43) ||
	(id == 0x64 && s != KERNEL_CHECKSUM_R64))
#endif
#endif
	fprintf(stderr,"Warning: KERNEL image may be corrupted. Sum: %d\n", s);


    if (Load_Sys_File (NULL, app_resources.basicName, basic_rom,
		       LOROM_BASE, LOROM_SIZE) < 0) {
	fprintf(stderr,"Couldn't load basic\n\n");
	return -1;
    }

    /* But fortunately not so many Basics */

    for (i = 0, s = 0; i < BASICROM_SIZE; i++) s += basic_rom[i];

#ifdef C128
    if (s != BASIC_CHECKSUM_85 && s != BASIC_CHECKSUM_86)
#else
    if (s != BASIC_CHECKSUM)
#endif
	fprintf(stderr,"Warning: BASIC image may be corrupted. Sum: %d\n", s);


#ifdef C128
    /* C128 Editor ROM */
    for (i = BASICROM_SIZE, s = 0; i < LOROM_SIZE; i++) s += basic_rom[i];
    if (id == 01 &&
	s != EDITOR_CHECKSUM_R01 &&
	s != EDITOR_CHECKSUM_R01swe && s != EDITOR_CHECKSUM_R01ger) {
	fprintf(stderr,"Warning: EDITOR image may be corrupted. Sum: %d\n", s);
	fprintf(stderr,"Check your Basic ROM\n");
    }
#endif

#ifdef C128
#if (RAMSIZE == 0x40000)
    memcpy(basic_rom + 0x6eb3 , inittab0, sizeof(inittab0));
#endif
#ifdef PATCH_ROM
    memcpy(basic_rom + mAdj + 0x3f10 , inittab1,  sizeof(inittab1));
#endif
#endif

/*    if (Load_Sys_File (NULL, "chargen", char_rom, CHAROM_BASE, HIROM_SIZE)) {
	fprintf(stderr,"You seem to have the character ROM too. I'll use it.\n\n");
    }*/


    if (module) {

#ifndef VIC20
	external_rom = allocate_mem(EXROM_SIZE<<1);

	if (Load_Sys_File(NULL, app_resources.exromName,
		((module == 2) ? external_rom + EXROM_SIZE : external_rom), 0,
		((module == 3) ? EXROM_SIZE<<1 : EXROM_SIZE))) {

	    fprintf(stderr,
		"Couldn't load module. Ignoring faulty virtual cartridge.\n");
	    module = 0;
	}
#else
	/* All modules are allocated in the open system RAM space */

	if (Load_Sys_File(NULL, app_resources.exromName,
		ram + (module * EXROM_SIZE), 0, EXROM_SIZE) < 0) {

	    fprintf(stderr,
		"Couldn't load module %d. Ignoring faulty virtual cartridge.\n",
		module);
	}
/*	if (Load_Sys_File(NULL, app_resources.exromName,
		ram + (module * EXROM_SIZE), 0,
		((module == 3) ? EXROM_SIZE<<1 : EXROM_SIZE)) < 0) {

	if (Load_Sys_File(NULL, app_resources.exromName,
		ram + (module * EXROM_SIZE), 0,
		((module == 3) ? EXROM_SIZE<<1 : EXROM_SIZE)) < 0) {
*/
#endif
    }
    return 0;  /* load ok */
}


/*
 * Memory is allocated and var is returned.
 */

BYTE   *allocate_mem(int size)
{
    BYTE *var;

    if (!size) {
	fprintf (stderr, "Allocate: No size.\n");
	exit (1);
    }

    var = (BYTE *)malloc(size * sizeof(BYTE));
    assert(var);

    return (var);
}


/*
 * Find and return filename with path.
 */

char   *create_name(char *path, char *name)
{
    static char buffer[256];
    char   *p;

    if (!path && !name) {
	printf ("No filename specified\n");
	return (NULL);
    }

	/* 'name' cannot be directory -- use 'path' instead */

    p = name;
    while (*p == '/' || *p == '~' || *p == '.')
	++p;

    if (!*p) {
	printf ("Invalid filename '%s'\n", name);
	return (NULL);
    }

    if (!name)
	name = "";

#if (defined(GEMDOS) || defined(MSDOS))
    if (*name == '/' || *name == '\\' || *name == '~' || *name == '.')
	sprintf(buffer, "%s", name);
    else
	sprintf(buffer,"%s\\%s",(path ? path : app_resources.directory), name);
#else
    if (*name == '/' || *name == '~' || *name == '.')
	sprintf(buffer, "%s", name);
    else
	sprintf(buffer, "%s/%s",(path ? path : app_resources.directory), name);
#endif

    return (buffer);
}


/*
 * This function loads the ROM images and system files and if reloc > 0,
 * removes possible load address in the beginning of the file.
 */

int     Load_Sys_File(char *path, char *name, BYTE *var, int reloc, int siz)
{
    FILE    *fp;
    char    *bufp;

    if ((bufp = create_name(path, name)) == NULL)
	return -2;

    if ((fp = fopen(bufp, READ)) == NULL) {
	perror(bufp);
	return -1;
    }

    if (reloc > 0) {
	if(((unsigned int)fgetc(fp) +256 *(unsigned int)fgetc(fp)) != reloc)
	    rewind(fp);
	else
	    printf("Skipping load address %04X on %s\n", reloc, name);
    }

    if (fread((char *)var, siz, 1, fp) != 1) {
	fprintf (stderr, "%s: Short file.\n", name);
	return -1;
    }

    (void) fclose(fp);

    return 0;  /* return ok */
}


/*
 * This function loads program files according to the reloc value or
 * the load address in the beginning on the file.
 * Note: The first two bytes in the beginning of file are always treated
 * as a load adress.
 */

int     LoadFile(char *path, char *name, BYTE *var, int reloc, int maxsiz)
{
    FILE    *fp;
    char    *bufp;
    int     length;
    int     b1, b2;	

    if ((bufp = create_name(path, name)) == NULL)
	return -2;

    if ((fp = fopen(bufp, READ)) == NULL) {
	perror(bufp);
	return -1;
    }

    b1 = fgetc(fp);
    b2 = fgetc(fp);

    if (reloc < 0) {
	reloc = LOHI ((BYTE)b1,(BYTE)b2);
    }
    maxsiz -= reloc;

    printf ("Loading '%s' from %04x, ", bufp, reloc);
    length = fread ((char *)var + reloc, 1, maxsiz, fp);
    printf ("%x bytes\n", length);
    (void) fclose(fp);
    return length;
}


ADDRESS get_text_begin(void)
{
    ADDRESS adr;

    adr = LOHI(ram[CBM_TXTTAB],ram[CBM_TXTTAB +1]);

    if (adr < 400 || ram[CBM_TXTTAB] != 1)
	adr = strtol(CBM_TXT_DEFAULT, NULL, hexflg ? 16 : 10);

    return (adr);
}


void    set_text_pointers(ADDRESS ls, ADDRESS end)
{
    ADDRESS le;
    BYTE   *p;

    /* Set end of load pointers like Kernel load. */

    ram[0xae] = end & 0xff;
    ram[0xaf] = (end >> 8) & 0xff;
#ifdef C128
    ram[CBM_VARTAB] = ram[CBM_ARYTAB] = ram[CBM_STREND] = 0;
    ram[CBM_VARTAB +1] = ram[CBM_ARYTAB +1] = ram[CBM_STREND +1] = 4;
#else
    ram[CBM_VARTAB] = ram[CBM_ARYTAB] = ram[CBM_STREND] = end & 0xff;
    ram[CBM_VARTAB +1] = ram[CBM_ARYTAB +1] = ram[CBM_STREND +1] =
	(end >> 8) & 0xff;
#endif

    /* Relink BASIC */

    if (ls == get_text_begin()) {
#ifdef DEBUG
	printf("Rechaining at %04X.\n", ls);
#endif
	while ((ram[ls] || ram[ls+1]) && ls < end) {
	    p = &ram[ls +4];
	    le = ls + 5;
	    while (*(p++))
		le++;
	    ram[ls] = le & 255;
	    ram[++ls] = (le>>8) & 255;
	    ls = le;
	}
    }
}


/*
 * This function is used to save any file
 * If reloc < 0, it is not included in the file.
 */

int     SaveFile(char *path, char *name, BYTE *var, int reloc, int siz)
{
    FILE    *fp;
    char    *bufp;
    int     length;

    if ((bufp = create_name(path, name)) == NULL)
	return -2;

    if ((fp = fopen(bufp, WRITE)) == NULL) {
	perror(bufp);
	return -1;
    }

    if (reloc >= 0) {
	fputc (LOWER (reloc), fp);
	fputc (UPPER (reloc), fp);
    }

    length = fwrite ((char *)var, 1, siz, fp);
    (void) fclose(fp);
    return length;
}


/* ------------------------------------------------------------------------- */

/* Set up C64 or C128 memory configuration */

void    init_memmaps(void)
{
    int     i;

    for (i = 0; i < 16; i++) {
	memwr[i] = ramwr;
	memrd[i] = ramrd;
    }
    memwr[0x00] = zerowr;

    set_video_hooks();

/*printf("InitMem PC = %04x\n", PC);*/
#ifdef C128

    /* C000 - FFFF */

    memwr[0x0f] = ramwr_hi;

    switch ((i = mmu[0]) & 0x30) {
      case 0x00:
	memrd[0xe] = memrd[0xf] = kernelrd;
	memrd[0xd] = charrd;   /* Chargen ROM */
	memrd[0xc] = basicrd;  /* Editor ROM */
dprintf("K-C-E ");
	break;
      case 0x10:
	if (module) {
	    memrd[0xe] = memrd[0xf] = exromrd; /* Int Func. ROM */
	    memrd[0xc] = memrd[0xd] = exromrd;
	}
dprintf("Fx-Hi ");
	break;
      case 0x20:
	if (module) {
	    memrd[0xe] = memrd[0xf] = exromrd; /* Ext Func. ROM */
	    memrd[0xc] = memrd[0xd] = exromrd;
	}
dprintf("EX-Hi ");
	break;
    }
    if (!(i & 1)) {
	memrd[0x0d] = iord;
	memwr[0x0d] = iowr;
dprintf("I/O ");
    }

    /* 8000 - BFFF */

    switch (i & 0x0C) {
      case 0x00:
	memrd[0xa] = memrd[0xb] = basicrd;
	memrd[0x8] = memrd[0x9] = basicrd;
dprintf("B-Hi ");
	break;
      case 0x04:
	if (module) {
	    memrd[0xa] = memrd[0xb] = exromrd; /* Int Func. ROM */
	    memrd[0x8] = memrd[0x9] = exromrd;
	}
dprintf("Fx-Lo ");
	break;
      case 0x08:
	if (module) {
	    memrd[0xa] = memrd[0xb] = exromrd; /* Ext Func. ROM */
	    memrd[0x8] = memrd[0x9] = exromrd;
	}
dprintf("EX-Lo ");
	break;
    }

    /* 4000 - 7FFF */

    if (!(i & 2)) {
	memrd[0x6] = memrd[0x7] = basicrd;
	memrd[0x4] = memrd[0x5] = basicrd;
dprintf("B-Lo ");
    }

/*printf("RAM %d\n", (mmu[0]>>6) & 3);*/

    /* Set up common RAM */

    ;

#else /* C128 */
#ifndef VIC20

/* Set up C64 memory configuration */

#ifdef REU
    if (reuflg)
	memwr[0x0f] = ramwr_hi;
#endif

    /*
     * Note: ULTIMAX mode (configurations 1, 5, 9 and D) is not supported
     */
    switch (ram[1] & 0x07) {
      case 7:  /* C E F */
	switch (module) {
	  case 1:
	    memrd[0x8] = memrd[0x9] = exromrd;

	  case 0:
	    memrd[0xa] = memrd[0xb] = basicrd;
	    break;

	  case 3:
	    memrd[0x8] = memrd[0x9] = exromrd;
	    memrd[0xa] = memrd[0xb] = exromrd;
	}

      case 6:  /* 4 6 7 */
	memrd[0xe] = memrd[0xf] = kernelrd;

	if (module == 2) {
	    memrd[0xa] = memrd[0xb] = exromrd;
	}

      case 5:  /* 8 A B */
	memrd[0x0d] = iord;
	memwr[0x0d] = iowr;

      case 4:  /* 0 2 3 */
	break;

      case 3:
	switch (module) {
	  case 1:
	    memrd[0x8] = memrd[0x9] = exromrd;

	  case 0:
	memrd[0xa] = memrd[0xb] = basicrd;
	    break;

	  case 3:
	    memrd[0x8] = memrd[0x9] = exromrd;
	    memrd[0xa] = memrd[0xb] = exromrd;
	}

      case 2:
	memrd[0xe] = memrd[0xf] = kernelrd;

	if (module == 2) {
	    memrd[0xa] = memrd[0xb] = exromrd;
	}
	memrd[0x0d] = charrd;
	break;

      case 1: /* the chargen is not accessible */
      case 0:
	break;
    }

#else /* VIC20 map */

    /*
     * The VIC20 memory map is quite straightforward, as there
     * isn't bank swithing involved there. Also, this function
     * needs to be called once at RESET only.
     */

    memrd[0xe] = memrd[0xf] = kernelrd;
    memrd[0xc] = memrd[0xd] = basicrd;
    memrd[0x8] = charrd;

    memrd[0x9] = iord;
    memwr[0x9] = iowr;

    switch (module) {
      case 1:
	memrd[0x2] = memrd[0x3] = exromrd;
	break;

      case 2:
	memrd[0x4] = memrd[0x5] = exromrd;
	break;

      case 3:
	memrd[0x6] = memrd[0x7] = exromrd;
	break;

      case 5:  /* A000 */
	memrd[0xa] = memrd[0xb] = exromrd;
    }


#endif /* VIC20 */
#endif /* C128 */


/* for convenience ... */

    memrd[0x10] = memrd[0];
    memwr[0x10] = memwr[0];
}


/* ------------------------------------------------------------------------- */

/*
 * I/O access functions
 * these are defined differently for each mode
 */

void iowr(ADDRESS address, BYTE byte)
{
#ifndef VIC20
    switch ((address & 0xf00) >> 8) {
	case 0:case 1:case 2:case 3:	/* VIC */
	store_vic(address & 0x3f, byte);
	return;

#ifdef C128
      case 5: /* MMU */
	store_mmu(address, byte);
	return;

      case 6: /* VDC */
	if (address & 1) {
#ifdef USE_VDC
	    store_vdc(vdc_ptr, byte);
#else
	    if (vdc_ptr < 37) {
		vdc[vdc_ptr] = byte;

		if (vdc_ptr == 0x1e) {   /* Copy/fill repetition count */
		    vdc[0x12] = LOAD(0xa3d);
		    vdc[0x13] = LOAD(0xa3c);
		}
	    }
#endif  /* USE_VDC */
	}
	else
	    vdc_ptr = byte;
	return;

#else
      case 5:
      case 6:
      case 7:
#endif	/* C128 */
      case 4: /* SID */
	sid[address & 0x1f] = byte;
	/* store sid */
	return;

      case 8:
      case 9:
      case 10:
      case 11:/* COLOR RAM */
	/* do not draw char unless its colour is changed */
	store_colorram(address & 0x3ff, byte);
	return;

      case 12:/* CIA 1 */
	store_cia1(address & 0xf, byte);
	return;

      case 13:/* CIA 2 */
	store_cia2(address & 0xf, byte);
	return;
#ifdef REU
      case 15: /* IO2 */
	if (reuflg)
	  store_reuc(address & 0xf, byte);
	return;
#endif
      default:
	return;
    }


#else  /* VIC20 I/O */

    switch ((address & 0xf00) >> 8) {
      case 0:	/* VIC */
	store_vic(address & 0xf, byte);
	return;

      case 1:/* VIA 1 */
	store_cia1(address & 0xf, byte);
	return;

	/* VIA 2 */
	store_cia2(address & 0xf, byte);
	return;

      case 4:
      case 5:
      case 6:
      case 7:/* COLOR RAM */
	/* do not draw char unless its colour is changed */
	store_colorram(address & 0x3ff, byte);
	return;

	/* 9800-9bff I/O1 */
	/* 9c00-9fff I/O1 */

      default:
	return;
    }

#endif
}


BYTE iord(ADDRESS address)
{
#ifndef VIC20

    switch ((address & 0xf00) >> 8) {
	case 0:case 1:case 2:case 3:	/* VIC */
	return (read_vic(address & 0x3f));

#ifdef C128
      case 5: /* MMU */
	return (((address &= 0xff) < 0x0b) ? mmu[address] : 0xff);

      case 6: /* VDC */
	if (address & 1) {
#ifdef USE_VDC
	    return (read_vdc(vdc_ptr));
#else
	    return ((vdc_ptr < 37) ? vdc[vdc_ptr] : 0xff);
#endif  /* USE_VDC */

	}
	else
	    return 0x9f; /* status */

#else
      case 5:
      case 6:
      case 7:  /* undefined in C128 */
#endif
      case 4: /* SID */
	if ((address & 0x1f) == 0x1b)
	    return (rand());
	return (sid[address & 0x1f]);

      case 8:
      case 9:
      case 10:
      case 11:/* COLOR RAM */
	return (read_colorram(address & 0x3ff));

      case 12:/* CIA 1  dc00 */
	return (read_cia1(address & 0xf));

      case 13:/* CIA 2  dd00 */
	return (read_cia2(address & 0xf));

      case 15: /* IO2   df00 */
#ifdef EMULATOR_ID
	if ((address & 0xff) > 0x9f) {
	    return (Emulator_ID[(address & 0xff) - 0xa0]);
	}
#endif
#ifdef REU
	if (reuflg)
	  return (read_reuc(address & 0xf));	/* else drop trough */
#endif
      default:
	return ((BYTE)(address >>8));		/* Empty slot */
    }


#else  /* VIC20 I/O */

    switch ((address & 0xf00) >> 8) {
      case 0:	/* VIC */
	return(read_vic(address & 0xf));

      case 1:
	return (read_cia1(address & 0xf));	/* VIA 1 */
	return (read_cia2(address & 0xf));	/* VIA 2 */

      case 4:
      case 5:
      case 6:
      case 7:/* COLOR RAM */
	return (read_colorram(address & 0x3ff));

	/* 9800-9bff I/O1 */
	/* 9c00-9fff I/O1 */

      default:
	return ((BYTE)(address >>8));
    }

#endif
}

/* ------------------------------------------------------------------------- */

/*
 * Memory access functions
 * these are common for each mode
 */

static void zerowr(ADDRESS address, BYTE byte)
{

#ifdef C128
    if (address == 1) {
	ram[1] = (0xff & ~ram[0]) | (byte & ram[0]);
    } else
	ram[RAMSHAR(address)] = byte;   /* by default only 1K common ram */

    /*if(address > 0x1ff)
	printf("RAMWR %X:%04X at PC = %04X  data %02x [ %02x %02x ]\n",
	       ((mmu[0] & 0xC0) >> 6), address, PC, byte,
	       ram[address], ram[address + 0x10000] );*/

#else
#ifndef VIC20
    static BYTE ram1;
    if (address == 1) {
	if ((ram[1] = (0xff & ~ram[0]) | (byte & ram[0])) != ram1) {
	    init_memmaps(); /* C64 */
	    ram1 = ram[1];
	}
    } else
#endif
	ram[address] = byte;
#endif
}


static void ramwr(ADDRESS address, BYTE byte)
{
    ram[RAMSHAR(address)] = byte;
}


#if (defined(REU) || defined(C128))
static void ramwr_hi(ADDRESS address, BYTE byte)
{
#ifdef C128
/*printf("BANK %x\n", RAMBANK());*/
    ram[RAMBANK() + address] = byte;

    if (address > 0xff00 && address <= 0xff04) { /* memory config */
	store_mmu(0, mmu[address & 15]);
    }
    else if (address == 0xff00) {
	store_mmu(0, byte);
#ifdef REU
	if (reuflg)	/* only when enabled */
	    reu_dma(-1);
#endif
    }

#else /* C128 */
#ifdef REU
    ram[address] = byte;
    if (address == 0xff00)	/* do not test every write */
      reu_dma(-1);
#endif
#endif
}
#endif


#ifdef C128
static void store_mmu(ADDRESS address, BYTE byte)
{
    static BYTE conf = 0;		/* bank15 -- default */

/* printf("%04X WR %04X: MMU[%x] = %02X\n", PC, address, address & 15, byte);
 */
    if ((address &= 0xff) < 0x0b) {
	mmu[address] = byte;

	switch (address) {
	  case 0:   /* mem config */
	    if (byte == conf)
		break;
	    conf = byte;

	/* MMU overrides everything, so CR must be always
	 * readable in the selected bank
	 */
	    kernal_rom[0x1f00] = byte;
	    ram[0xff00] = ram[0x1ff00] = byte;
#if (RAMSIZE == 0x40000)
	    ram[0x2ff00] = ram[0x3ff00] = byte;
#endif
	    /*ram[RAMSHAR(0xff00)] = byte;*/


	  case 6:   /* ram config */
	    init_memmaps(); /* C128 */
	    break;

	  case 5:   /* operating mode */
	    byte = (byte & 0x7f) | 0x30; /* read-only bits */
	    if ((byte & 0x41) != 1)
		fprintf(stderr,
		       "D505 %02X - Attempted accessing unimplemented mode.\n",
			byte);
	}
    }
    return;
}
#endif


static BYTE ramrd(ADDRESS address)
{
    return (ram[RAMSHAR(address)]);
}


static BYTE basicrd(ADDRESS address)
{
    return (basic_rom[BASIC_ADDRESS(address)]);
}


static BYTE kernelrd(ADDRESS address)
{
    return (kernal_rom[address & 0x1fff]);
}


static BYTE exromrd(ADDRESS address)
{
    /*printf("Exromrd %04X\n", address);*/
    return (external_rom[address & 0x3fff]);
}


static BYTE charrd(ADDRESS address)
{
    return (char_rom[address & 0xfff]);
}
