/*
 * $Id: memory.h,v 1.3 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains the Commodore OS constants and such.
 *
 * Included are:
 *	o memory banking
 *	o BASIC constants
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 *
 * $Log: memory.h,v $
 * Revision 1.3  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * C128 ROM checksums done better way.
 *
 * Revision 1.2  1994/12/12  16:59:44  jopi
 * 16-bit portability
 *
 *
 *
 */

#ifndef X64_MEMORY_H
#define X64_MEMORY_H


#include "config.h"
#include "types.h"
#include "file.h"


/* INTERRUPTS */

/* System Memory Constants and OS Constants for Monitor
 * (The latter are only for convenience, since x64 boots from FFFC)
 * Also checksums for all *known* revisions are provided, except for
 * those third party turbo systems.
 */

#define HIROM_SIZE		0x2000   /* Kernel */

#ifdef C128
#define RAMSIZE			0x20000  /* x128 supports 0x40000 too */
#define BASICROM_SIZE		0x8000
#define EDITROM_SIZE		0x1000
#define LOROM_SIZE	0x9000 /* load Basic (LO,HI) and Editor at once */
#define LOROM_MASK		0x3FFF /* 16K + 16K + 4K */

#define EXROM_SIZE	0x4000  /* C128 function rom is 2*16 KB slots */

#define CHAROM_BASE		0xD000
#define LOROM_BASE		0x4000
#define BASIC_ADDRESS(a) 	((a) - LOROM_BASE)

#define BASIC_CHECKSUM_85	38592
#define BASIC_CHECKSUM_86	2496
#define EDITOR_CHECKSUM_R01	56682
#define EDITOR_CHECKSUM_R01swe	9364
#define EDITOR_CHECKSUM_R01ger	9619
#define KERNEL_CHECKSUM_R01	22353
#define KERNEL_CHECKSUM_R01swe	24139
#define KERNEL_CHECKSUM_R01ger	22098

#define RESTART_BASIC		0x4000  /* or 4003 */

/*
 * BASIC Pointers
 */
#define CBM_TXTTAB		0x2D	/* Basic Text */
#define CBM_VARTAB		0x2F	/* Basic Variables */
#define CBM_ARYTAB		0x31	/* Basic Arrays */
#define CBM_STREND		0x33	/* End of Arrays +1 */
#define CBM_FRETOP		0x35	/* Bottom of String Storage */
#define CBM_TXT_DEFAULT		"0x1C01"


#else  /* not C128 */

#ifdef VIC20
#define RAMSIZE			0x10000	/* for easier module handling... */
#define BASICROM_SIZE		0x2000
#define LOROM_SIZE		BASICROM_SIZE
#define LOROM_MASK		0x1FFF
#define CHAROM_BASE		0x8000
#define LOROM_BASE		0xC000

#define BASIC_CHECKSUM		33073
#define KERNEL_CHECKSUM		38203

#define RESTART_BASIC		0x0
#define CBM_TXT_DEFAULT		"0x1201"

#else  /* C64 */
#define RAMSIZE			0x10000
#define BASICROM_SIZE		0x2000
#define LOROM_SIZE		BASICROM_SIZE
#define LOROM_MASK		0x1FFF
#define CHAROM_BASE		0xD000
#define LOROM_BASE		0xA000

#define BASIC_CHECKSUM		15702
#define KERNEL_CHECKSUM_R00	50955
#define KERNEL_CHECKSUM_R03	50954
#define KERNEL_CHECKSUM_R03swe	50633
#define KERNEL_CHECKSUM_R43	50955
#define KERNEL_CHECKSUM_R64	49680

#define RESTART_BASIC		0xA474
#define CBM_TXT_DEFAULT		"0x0801"

/*#define RESTART_COLD		0xFCE2*/
/*#define RESTART_WARM		0xFCFE*/

#endif  /* C64 */


#define EXROM_SIZE		0x2000  /* C64/VIC20 module is 2*8 KB slots */
#define BASIC_ADDRESS(a) 	((a) & LOROM_MASK)

/*
 * BASIC Pointers
 */
#define CBM_TXTTAB		0x2B	/* Basic Text */
#define CBM_VARTAB		0x2D	/* Basic Variables */
#define CBM_ARYTAB		0x2F	/* Basic Arrays */
#define CBM_STREND		0x31	/* End of Arrays +1 */
#define CBM_FRETOP		0x33	/* Bottom of String Storage */

#endif  /* C64 & VIC20 */


/*
 * Functions
 */

extern BYTE    load_rom( ADDRESS a );
extern void    save_rom( ADDRESS a, BYTE x );

extern BYTE   *allocate_mem ( int size );
extern char   *create_name ( char *path, char *name );
extern int     Load_Sys_File ( char *path, char *name, BYTE *var, int reloc, int siz );
extern int     LoadFile ( char *path, char *name, BYTE *var, int reloc, int maxsiz );
extern int     SaveFile ( char *path, char *name, BYTE *var, int reloc, int siz );

extern void    store_vic ( ADDRESS addr, BYTE byte );
extern void    store_colorram ( ADDRESS addr, BYTE byte );
extern BYTE    read_colorram ( ADDRESS addr );
 

/* ram expansion */
#ifdef REU

extern int reuflg;

extern int     reset_reuc ( char *, ADDR_T );
extern void    close_reuc ( char * );
extern BYTE    read_reuc ( ADDRESS );
extern void    store_reuc ( ADDRESS, BYTE );
extern void    reu_dma ( int );

#endif  /* REU */


/* rom patcher */

#ifdef PATCH_ROM

extern int patch_rom ( char * );

#endif  /* PATCH_ROM */


#endif  /* X64_MEMORY_H */
