/*
 * $Id: serial.c,v 1.7 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Serial interface control routines.
 *
 *
 * Written by
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *
 *
 * $Log: serial.c,v $
 * Revision 1.7  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * New serial device control.
 *
 * Revision 1.6  1994/12/12  16:59:44  jopi
 * Run-time device control.
 *
 * Revision 1.5  1994/06/16  17:19:26  jopi
 * C128 traps. Support for multiple drives etc.
 * Code reorganized and cleaned up.
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * *** empty log message ***
 *
 * Revision 1.3  93/06/21  13:39:21  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

/*
 * serial.c -- Serial device interface.
 * The serial device interface consists of traps in serial access routines
 * TALK, ACPTR, Send Data, and Get Data, so that actual operation can be
 * controlled by C routines.  Serial.c implements the I/F control whereas
 * the required acknowledge is supplied with the peripheral emulators.
 * To support turboloaders, it would be necessary to emulate the serial port
 * itself. That is much slower though. :(
 *
 */

#include <stdio.h>

#include "cpu.h"
#include "macro.h"
#include "extern.h"
#include "serial.h"
#include "config.h"

#include "printer.h"
#include "drive.h"


#ifdef C128
#define TRAP_LISTEN_RTS		0xE5BA
#define TRAP_LISTEN_NODEV	0xE428
#else
#define TRAP_LISTEN_RTS		0xEDAB
#define TRAP_LISTEN_NODEV	0xEDAD
#endif

#define SET_ST(b)	STORE(0x90, (LOAD(0x90) | b))


/*
 * Initialized serialdevices
 */
#define MAXDEVICES	16
static serial_t serialdevices[MAXDEVICES];

/*
 * Store name here for serial-open
 */
#define SERIAL_NAMELENGTH 255

static BYTE SerialBuffer[SERIAL_NAMELENGTH +1];
static int SerialPtr;


/*
 * On which channel did listen happen to ?
 */
static int TrapDevice;
static int TrapSecondary;


/*
 * Functions
 */

static void serialattention();
static void serialsendbyte();
static void serialreceivebyte();
static void trap_serial_ready( void );

static int fn();

/*
 * Extern functions
 */

extern void detach_print ( PRINTER *printer );
extern int  attach_print ( PRINTER *printer, char *name, int mode );
extern int  attach_floppy_image ( DRIVE *floppy, char *name, int mode );
extern void detach_floppy_image ( DRIVE *floppy );


/*
 * Call this if device is not attached
 *  -128 == device not present
 */
static int fn()
{
    return (0x80);
}


/* ------------------------------------------------------------------------- */

/*
 * Handle Serial Bus Commands under Attention.
 */

static int  serialcommand()
{
    serial_t *p;
    int     channel;
    int     i, st = 0;

#ifdef DEBUG_SERIAL
    printf("***SerialCommand %02x %02x %02x  BUF(%d) '%s'\n\n",
	   TrapDevice, TrapSecondary, ram[0x95],  SerialPtr, SerialBuffer);
#endif

    /*
     * which device ?
     */
    p = &serialdevices[TrapDevice & 0x0f];
    channel = TrapSecondary & 0x0f;

    switch (TrapSecondary & 0xf0) {
	/*
	 * Open Channel
	 */
      case 0x60:
	if (!p -> isopen) {
	    p -> isopen = 1;
	    st = (*(p ->openf))(p ->info, NULL, 0, channel);
	    for (i = 0; i < SerialPtr; i++)
		(*(p -> putf))(p -> info, SerialBuffer[i], channel);
	    SerialPtr = 0;
	}
	if (p -> flushf)
	    (*(p -> flushf))(p -> info, channel);
	break;

	/*
	 * Close File
	 */
      case 0xE0:
	p -> isopen = 0;
	st = (*(p -> closef))(p -> info, channel);
	break;

	/*
	 * Open File
	 */
      case 0xF0:
	if (!p -> isopen) {
	    p -> isopen = 1;
	    SerialBuffer[SerialPtr] = 0;
	    st = (*(p -> openf))(p -> info, (char *)SerialBuffer, SerialPtr,
				 channel);
	    SerialPtr = 0;

	    if (st) {
		p -> isopen = 0;
		(*(p -> closef))(p -> info, channel);

		fprintf(stderr, "Serial: Cannot open file. Status %02x\n", st);
	    }
	}
	if (p -> flushf)
	    (*(p -> flushf))(p -> info, channel);
	break;

      default:
	printf("Unknown command %02X\n\n", TrapSecondary & 0xff);
    }
    return (st);
}


/* ------------------------------------------------------------------------- */

/*
 * Command Serial Bus to TALK, LISTEN, UNTALK, or UNLISTEN, and
 * send the Secondary Address to Serial Bus under Attention.
 */

static void serialattention()
{
    int     b;
    int     st;

    /*
     * Which Secondary Address ?
     */
    b = LOAD(0x95);		/* BSOUR - character for serial bus */

#ifdef DEBUG_SERIAL
    printf("SerialSaListen(%02x)\n", b);
#endif

    if (b == 0x3f && (TrapSecondary & 0xf0) == 0xf0) {
	st = serialcommand();
	SET_ST( st);
    }
    else switch (b & 0xf0) {
      case 0x20:
      case 0x40:
	TrapDevice = b;
	break;

      case 0x60:
      case 0xe0:
	TrapSecondary = b;
	st = serialcommand();
	SET_ST( st);
	break;

      case 0xf0:	/* Open File needs the filename first */
	serialdevices[TrapDevice & 0x0f].isopen = 0;
	TrapSecondary = b;
#ifdef DEBUG_SERIAL
	printf("close output.\n");
#endif
	break;
    }

    if (!(serialdevices[TrapDevice & 0x0f].inuse))
	SET_ST( 0x80);

    /*
     * Set PC to 'RTS'
     */
    SET_CARRY(0);
    PC = TRAP_LISTEN_RTS;
}


static void serialsendbyte()
{
    int     data, st;
    serial_t *p;

    /*
     * Get data to send from address 0x95
     */
    data = LOAD(0x95);		/* BSOUR - character for serial bus */

    p = &serialdevices[TrapDevice & 0x0f];

    if (p -> inuse) {
	if (!p -> isopen) {

#ifdef DEBUG_SERIAL
	    printf("SerialSendByte[%2d] = %02x\n", SerialPtr, data);
#endif
	    /* Store name here */
	    if (SerialPtr < SERIAL_NAMELENGTH)
		SerialBuffer[SerialPtr++] = data;
	} else {

#ifdef DEBUG_SERIAL_1
	    printf("SerialSendByte(%02x)\n", data);
#endif
	    /* Send to device */
	    st = (*(p -> putf))(p -> info, data, TrapSecondary & 0x0f);
	    SET_ST( st);
	}
    }
    else {	/* Not present */
	SET_ST(0x83);
    }

    /*
     * Set PC to RTS
     */

    SET_CARRY(0);
    PC = TRAP_LISTEN_RTS;
}

static void serialreceivebyte()
{
    int     st = 0;
    BYTE    data;
    serial_t *p;

    p = &serialdevices[TrapDevice & 0x0f];
    st = (*(p -> getf))(p -> info, &data, TrapSecondary & 0x0f);

#ifdef DEBUG_SERIAL_1
    printf("SerialReceiveByte(%02x)\n", data);
#endif

    /*
     * Set up serial success / data
     */
    if (st) SET_ST(st);
    STORE(0xA4, data);


    /*
     * Set registers (PC, AC and CARRY) like the ML routine does
     */
    AC = data;
    SET_CARRY(0);

    PC = TRAP_LISTEN_RTS;
}


/*
 * Kernal loops serial-port (0xdd00) to see when serial is ready. Fake it.
 *  EEA9  Get serial data and clk in (TKSA subroutine)
 */

static void trap_serial_ready(void)
{
    AC = 0x01;
    SET_SIGN(AC);
    SET_ZERO(AC);
#ifdef C128
    PC = 0xE572;	/* C128: RTS */
#else
    PC = 0xEEB2;	/* C64: RTS */
#endif
}


/* ------------------------------------------------------------------------- */

/*
 * Initialize serial module. Set up traps.
 */

static trap_t init_traps[] = {
#ifndef C128
/* C64 Traps */
    {
	"SerialListen",
	0xED24,
	{0x20, 0x97, 0xEE},
	serialattention
    },
    {
	"SerialSaListen",
	0xED36,
	{0x78, 0x20, 0x8E},
	serialattention
    },
    {
	"SerialSendByte",
	0xED40,
	{0x78, 0x20, 0x97},
	serialsendbyte
    },
    {
	"SerialReceiveByte",
	0xEE13,
	{0x78, 0xA9, 0x00},
	serialreceivebyte
    },
    {
	"Serial ready",
	0xEEA9,
	{0xAD, 0x00, 0xDD},
	trap_serial_ready
    }

#else
/* C128 Traps */
    {
	"SerialListen",
	0xE355,
	{0x20, 0x73, 0xE5},
	serialattention
    },
    {
	"SerialSaListen",
	0xE37C,
	{0x20, 0x73, 0xE5},
	serialattention
    },
    {
	"SerialSendByte",
	0xE38C,
	{0x20, 0x73, 0xE5},
	serialsendbyte
    },
    {
	"SerialReceiveByte",
	0xE43E,
	{0x20, 0x73, 0xE5},
	serialreceivebyte
    },
    {
	"Serial ready",
	0xE569,
	{0xAD, 0x00, 0xDD},
	trap_serial_ready
    },
    {
	"serial ready",
	0xE4F5,
	{0xAD, 0x00, 0xDD},
	trap_serial_ready
    }
#endif
};


/* ------------------------------------------------------------------------- */

void    initialize_serial()
{
    int     i;

    /*
     * Which traps to initialize
     */

    set_trap(&init_traps[0]);
    set_trap(&init_traps[1]);
    set_trap(&init_traps[2]);
    set_trap(&init_traps[3]);

    set_trap(&init_traps[4]);
#ifdef C128
    set_trap(&init_traps[5]);
#endif

    /*
     * Clear serial device functions
     */
    for (i = 0; i < MAXDEVICES; i++) {
	serial_t *p;

	p = &serialdevices[i];

	p -> getf   = (int (*)(void *, BYTE *, int))fn;
	p -> putf   = (int (*)(void *, BYTE, int))fn;
	p -> openf  = (int (*)(void *, char *, int , int))fn;
	p -> closef = (int (*)(void *, int))fn;
	p -> flushf = (void (*)(void *, int))NULL;
    }
}


int     attach_serial_device(int device, char *var, char *name, 
			int (*getf)(void *, BYTE * , int), 
			int (*putf)(void *, BYTE , int), 
			int (*openf)(void *, char *, int , int), 
			int (*closef)(void *, int), 
			void (*flushf)(void *, int))
{
    serial_t *p;

    if (device < 0 || device >= MAXDEVICES)
	return 1;

    p = &serialdevices[device];

    if (p -> inuse) {
	printf("warning. serial device %d (%s) in use\n", device, name);
	return 1;
    }
    p -> getf  = getf;
    p -> putf  = putf;
    p -> openf = openf;
    p -> closef = closef;
    p -> flushf = flushf;

    p -> inuse = 1;
    p -> info  = var; /* run-time data storage */
    p -> name  = xstrdup(name);
    return 0;
}


/*
 * attach a file for use to specified peripheral. The device
 * numbed is searched in filename if device # == -1.
 */

int  serial_select_file(int type, int number, char *file)
{
    serial_t *p;

    if (number < 0 && ( (number = find_devno(-1, file)) < 0) && type ) {
	if (type & DT_PRINTER)
	    number = 4;
	else if (type & DT_DISK)
	    number = 8;
    }
    if (number < 0 || number >= MAXDEVICES) {
	fprintf (stderr, "\nIllegal device number %d.\n", number);
	return -1;
    }

    p = &serialdevices[number];

    /*
     * Create a new device if it isn't initialised, and a type is given.
     */

    if (!p || !(p->inuse)) {
	if (type) {
	    switch (number) {
	      case 1:
	      case 2:
		fprintf (stderr, "\nUnsupported device, number %d.\n", number);
		return -1;
	      case 4:
	      case 5:
		initialize_printer(number, app_resources.PrinterLang,
				   app_resources.Locale );
		break;
	      case 8:
	      case 9:
	      case 10:
	      case 11:
		initialize_1541(number, type);
		break;
	      default:
		fprintf (stderr, "\nIllegal device number %d.\n", number);
		return -1;
	    }
	}
	else {
	    printf("\nSorry, serial device %d not available.\n", number);
	    return 1;
	}
    } /* if !p */


 /* should be based on p -> info.type */
   if (number >= 4 && number <= 5 ) {
	attach_print((PRINTER *)p -> info, file, 0);
    }
    else if (number >= 8 && number <= 11 ) {
	return (attach_floppy_image((DRIVE *)p -> info, file, 0));
    }


    return (0);
}


/*
 * Detach and kill serial devices.
 * detach all (shutdown) if device # == -1.
 */

int  remove_serial(int number)
{
    serial_t *p;
    int  i;

    if (number < 0) {
	for (i = 0; i < MAXDEVICES; i++) {

	    p = &serialdevices[i];

	    if (p && p->inuse) {
		if (number == -2) {	/* QUERY mode */
		    printf("Unit #%d: %s\n.", i, p -> name);
		}
		else {
#if (defined(DEBUG) || defined(DEBUG_SERIAL))
		    printf("remove unit #%d.\n", i);
#endif
		    remove_serial(i);
		}
	    }
	}
    }
    else {
	if (number < 0 || number >= MAXDEVICES) {
	    fprintf (stderr, "\nIllegal device number %d.\n", number);
	    return -1;
	}

	p = &serialdevices[number];

	if (!p || !(p->inuse)) {
	    printf("I'm not here.  ;)\n");
	}
	else
	    switch (number) {  /* should be based on actual type ... */
	      case 4:
	      case 5:
		detach_print((PRINTER *)p -> info);
		break;
	      case 8:
	      case 9:
	      case 10:
	      case 11:
		detach_floppy_image((DRIVE *)p -> info);
		break;

	      default:
		printf ("Orphan device #%d\n", number);
	    }

    } /* else */

    return (0);
}
