/*
 * $Id: vic_new.c,v 1.2 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * VIC-II chip emulation. VIC is (to be) implemented using generic
 * high-speed drawing routines. This makes possible to use different
 * display drivers for each terminal type.
 *
 *
 * Written by
 *   Hannu Helminen  (dm@stekt.oulu.fi)
 *
 *
 * $Log: vic_new.c,v $
 * Revision 1.2  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Compability.
 *
 * Revision 1.1  1994/12/12  16:59:44  jopi
 * Initial revision
 *
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "vmachine.h"
#include "extern.h"
#include "vic.h"

static VIC *vic = 0;

#define CollideCharMask		0x10
#define CollideSpriteMask	0xe0


void  DrawNormalChar( char *s, BYTE *text, BYTE *colorram, BYTE *font,
	int y, int xsize, int color0, int color1, int color2 );
void  DrawNormalHires( char *s, BYTE *text, BYTE *colorram, BYTE *font,
	int y, int xsize, int color0, int color1, int color2, int counter );
void  DrawMulticolorChar( char *s, BYTE *text, BYTE *colorram, BYTE *font,
	int off, int y, int xsize, int color0, int color1, int color2 );

VIC  *VicInit( void );
void  VicDrawSlab ( VIC *vic, int xx1, int y1, int xx2, int y2 );
void  VicVerticalBlank ( VIC *vic );
char *get_vic_bitmap ( void );


/*********************************************************************
* Various high-speed drawing routines
*********************************************************************/

/* A 10-bit wrapping counter */
#define INC10(ttt) \
  ((ttt) = ((ttt) + 1) & 0x3ff)


void DrawNormalChar(
  char *s, BYTE *text, BYTE *colorram, BYTE *font,
  int y, int xsize, int color0, int color1, int color2
) {
  font += y;
  while (--xsize >= 0) {
    int color = (*colorram++ & 0x0f) | CollideCharMask;
    int p = font[ *text++ << 3 ];
    if (p & 0x80) s[0] = color; else s[0] = color0;
    if (p & 0x40) s[1] = color; else s[1] = color0;
    if (p & 0x20) s[2] = color; else s[2] = color0;
    if (p & 0x10) s[3] = color; else s[3] = color0;
    if (p & 0x08) s[4] = color; else s[4] = color0;
    if (p & 0x04) s[5] = color; else s[5] = color0;
    if (p & 0x02) s[6] = color; else s[6] = color0;
    if (p & 0x01) s[7] = color; else s[7] = color0;
    s += 8;
  }
}


void DrawNormalHires(
  char *s, BYTE *text, BYTE *colorram, BYTE *font,
  int y, int xsize, int color0, int color1, int color2, int counter
) {
  font += y;
  counter *= 8;
  while (--xsize >= 0) {
    int colorB = *text & 0x0f;
    int colorF = *text++ >> 3 | CollideCharMask;
    int p = font[ counter += 8 ];
    if (p & 0x80) s[0] = colorF; else s[0] = colorB;
    if (p & 0x40) s[1] = colorF; else s[1] = colorB;
    if (p & 0x20) s[2] = colorF; else s[2] = colorB;
    if (p & 0x10) s[3] = colorF; else s[3] = colorB;
    if (p & 0x08) s[4] = colorF; else s[4] = colorB;
    if (p & 0x04) s[5] = colorF; else s[5] = colorB;
    if (p & 0x02) s[6] = colorF; else s[6] = colorB;
    if (p & 0x01) s[7] = colorF; else s[7] = colorB;
    s += 8;
  }
}


void DrawMulticolorChar(
  char *s, BYTE *text, BYTE *colorram, BYTE *font, int off,
  int y, int xsize, int color0, int color1, int color2
) {
  font += y;
  color2 |= CollideCharMask;
  while (--xsize >= 0) {
    int color = (*colorram++ & 0x0f) | CollideCharMask;
    int p = font[ *text++ << 3 ];
    if (p & 0x80) {
	if (p & 0x40) {s[0] = color;  s[1] = color;}
      else          {s[0] = color2; s[1] = color2;}
    } else {
      if (p & 0x40) {s[0] = color1; s[1] = color1;}
      else          {s[0] = color0; s[1] = color0;}
    }

    if (p & 0x20) {
      if (p & 0x10) {s[2] = color;  s[3] = color;}
      else          {s[2] = color2; s[3] = color2;}
    } else {
      if (p & 0x10) {s[2] = color1; s[3] = color1;}
      else          {s[2] = color0; s[3] = color0;}
    }

    if (p & 0x08) {
      if (p & 0x04) {s[4] = color;  s[5] = color;}
      else          {s[4] = color2; s[5] = color2;}
    } else {
      if (p & 0x04) {s[4] = color1; s[5] = color1;}
      else          {s[4] = color0; s[5] = color0;}
    }
    if (p & 0x02) {
      if (p & 0x01) {s[6] = color;  s[7] = color;}
      else          {s[6] = color2; s[7] = color2;}
    } else {
      if (p & 0x01) {s[6] = color1; s[7] = color1;}
      else          {s[6] = color0; s[7] = color0;}
    }
    s += 8;
  }
}


/*********************************************************************
* VIC structure
*********************************************************************/

VIC *VicInit()
{
  VIC *vic = (VIC *)malloc(sizeof(VIC));
  ASSERT(vic);

  vic->screen = (char *)malloc(VicXTotal * VicYTotal);
  memset(vic->screen, 0, VicXTotal * VicYTotal);
  vic->colorRam = (BYTE *)malloc(VicXTotal * VicYTotal * sizeof(BYTE));
  vic->bank = 0;
  return(vic);
}


void VicDrawSlab(VIC *vic, int xx1, int y1, int xx2, int y2)
{
  int sy;
  int x1, x2;
  BYTE *reg = vic->reg;
  void (*func)() = 0;

  enum {EXTENDCOLOR = 0x01, BITMAP = 0x02, MULTICOLOR = 0x04} mode = 0;

  unsigned bankbase = vic->bank * 0x4000;
  BYTE *screenmem = ram + bankbase + (vic->reg[0x18] & 0xf0) * 0x40;
  BYTE *charmem   = ram + bankbase + (vic->reg[0x18] & 0x0e) * 0x400; 

  if (charmem == ram + 0x1000 || charmem == ram + 0x9000) {
      charmem = &char_rom[0x0000];
  } else if (charmem == ram + 0x1800 || charmem == ram + 0x9800) {
      charmem = &char_rom[0x0800];
  }

  if (reg[0x11] & 0x40) {
    mode |= EXTENDCOLOR;
  }
  if (reg[0x11] & 0x20) {
    mode |= BITMAP;
  }
  if (reg[0x16] & 0x10) {
    mode |= MULTICOLOR;
  }

  switch(mode) {
    case 0: func = DrawNormalChar; break;
    case MULTICOLOR: func = DrawMulticolorChar; break;
    case BITMAP: func = DrawNormalHires; break;
    default: func = 0;
  }
  if (!func)
    return;

  for (sy = y1, x1 = xx1; sy <= y2; sy++, x1 = 0) {
    int yr = ((sy + 5) & 0x07);
    if (sy == y2)
      x2 = xx2;
    /* y row (0-7) inside a character */

    /* check if we should be doing a bad line (ie. to read the
     * character codes from screen memory */

#define FirstBadlineRow (VicYExp - 3 - 1)
#define LastBadlineRow (FirstBadlineRow + VicYSize - 1)

    if (sy >= FirstBadlineRow && sy <= LastBadlineRow && yr == 7) {
#define CharPrefetch 2 /* Please check this! */
      int xc1 = x1 - (VicXExp - CharPrefetch);
      int xc2 = x2 - (VicXExp - CharPrefetch);
      int count = vic->charCount;
      int x;
      if (xc1 < 0) xc1 = 0;
      if (xc2 >= 40) xc2 = 39;
      for (x = xc1; x <= xc2; x++) {
        vic->charBuf[x] = screenmem[count];
        vic->colorBuf[x] = vic->colorRam[count];
        INC10(count);
      }
      vic->charCount = count;
    }

    func(vic->screen + VicXExp + VicXTotal * sy,
         &vic->charBuf, &vic->colorBuf, charmem,
         yr, 0x28,
         vic->reg[0x21], vic->reg[0x22], vic->reg[0x23],
         vic->charCount-40);
  }
}

void VicVerticalBlank(VIC *vic)
{
  vic->charCount = 0;
}


/*********************************************************************
* interfaces
*********************************************************************/


void store_vic(ADDRESS addr, BYTE byte) {
  vic->reg[addr] = byte;

#if OLD_VIC
void    store_vic(ADDRESS addr, BYTE byte)
{
    vic[addr] = byte;
    if (debugflg)
	printf("store vic[%x] %x\n", addr, (int) byte);

    switch (addr) {
      case 0:
      case 1:
      case 2:
      case 3:
      case 4:
      case 5:
      case 6:
      case 7:
      case 8:
      case 9:
      case 10:
      case 11:
      case 12:
      case 13:
      case 14:
      case 15:/* SPRITE COORDINATES */
	addr >>= 1;
	if (vic[0x15] & pow2[addr])
	    DRAW_SPRITE(addr);
	return;
      case 16:
	{
	    static  last_x_msb;
	    BYTE    t = last_x_msb ^ byte;
	    int     i = 0;
	    last_x_msb = byte;
	    while (t) {
		if (t & 1)
		    DRAW_SPRITE(i);
		t >>= 1;
		i++;
	    }
	    return;
	}

      case 17:/* $D011 */
	screen_init();
	update_raster(1);
	return;

      case 18:/* $D012 raster line compare */
	update_raster(1);
	return;

      case 19:
      case 20:/* Light Pen X,Y */
	return;

      case 21:/* $D015 SPRITE enable */
	{
	    static  last_spr_mask;
	    BYTE    t = last_spr_mask ^ byte;
	    int     i = 0;
	    last_spr_mask = byte;

	    while (t) {
		if (t & 1) {
		    if (byte & 1) {
			x_change_sprite
			    (i, calc_sprite_addr(scrbase[0x3f8 + i]),
			    vic[0x1d] & pow2[i],
			    vic[0x17] & pow2[i]);
			x_draw_sprite(i);
			DRAW_SPRITE(i);
		    } else
			x_undraw_sprite(i);
		}
		t >>= 1;
		byte >>= 1;
		i++;
	    }
	    return;
	}

      case 22:/* $D016 */
	screen_init();	/* possible multicolor mode */
	return;

      case 23:/* $D017 sprite expand y */
	{
	    static  last_exp_y;
	    BYTE    t = last_exp_y ^ byte;
	    int     i = 0;
	    last_exp_y = byte;

	    while (t) {
		if (t & 1)
		    x_change_sprite(i, calc_sprite_addr(scrbase[0x3f8 + i]),
			vic[0x1d] & pow2[i], vic[0x17] & pow2[i]);
		t >>= 1;
		i++;
	    }
	}
	return;

      case 24:/* $D018 video and char matrix base address */
	screen_init();
	return;

      case 25:/* $D019 IRQ flag register */
	return;
      case 26:/* $D01A IRQ mask register */
	/* light pen interrupt not implemented */
	if (byte & 8)
	    int_clk[I_LIGHT_PEN] = 0;
	else
	    int_clk[I_LIGHT_PEN] = 0;

	/* sprite to sprite collision not yet implemented */
	if (byte & 4)
	    int_clk[I_SS_COL] = 0;
	else
	    int_clk[I_SS_COL] = 0;

	/* sprite to background collision not yet implemented */
	if (byte & 2)
	    int_clk[I_SB_COL] = 0;
	else
	    int_clk[I_SB_COL] = 0;

	/* raster interrupts */
	if (byte & 1)
	    update_raster(1);
	else
	    int_clk[I_RASTER] = 0;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	return;

      case 27:/* $D01B sprite to BG priority */
	return;
      case 28:/* $D01C sprite multicolor */
	return;
      case 29:/* $D01D sprite expand x */
	{
	    static  last_exp_x;
	    BYTE    t = last_exp_x ^ byte;
	    int     i = 0;
	    last_exp_x = byte;

	    while (t) {
		if (t & 1)
		    x_change_sprite(i, calc_sprite_addr(scrbase[0x3f8 + i]),
			vic[0x1d] & pow2[i], vic[0x17] & pow2[i]);
		t >>= 1;
		i++;
	    }
	}
	return;

      case 30:/* $D01E sprite collision */
	return;
      case 31:/* $D01F sprite to BG collision */
	return;

	/* colors */
      case 32:/* $D020 BORDER COLOR */
	x_change_color(C_BORDER, byte);
	return;
      case 33:/* $D021 BACKGROUND 0 */
	x_change_color(C_BG0, byte);
	return;
      case 34:/* $D022 BACKGROUND 1 */
	x_change_color(C_BG1, byte);
	return;
      case 35:/* $D023 BACKGROUND 2 */
	x_change_color(C_BG2, byte);
	return;
      case 36:/* $D024 BACKGROUND 3 */
	x_change_color(C_BG3, byte);
	return;
      case 37:/* sprite multicolor register 0 */
	x_change_color(C_SPRMUL0, byte);
	return;
      case 38:/* sprite multicolor register 1 */
	x_change_color(C_SPRMUL1, byte);
	return;
      case 39:/* sprite colours */
	x_change_color(C_SPR0, byte);
	return;
      case 40:
	x_change_color(C_SPR1, byte);
	return;
      case 41:
	x_change_color(C_SPR2, byte);
	return;
      case 42:
	x_change_color(C_SPR3, byte);
	return;
      case 43:
	x_change_color(C_SPR4, byte);
	return;
      case 44:
	x_change_color(C_SPR5, byte);
	return;
      case 45:
	x_change_color(C_SPR6, byte);
	return;
      case 46:/* $D02E */
	x_change_color(C_SPR7, byte);
	return;
#ifdef C128
      case 47:/* $D02F additional KBD rows */
	return;
      case 48:/* $D030 test, fast */
	return;
#endif
    }
}
#endif  /* OLD_VIC */
}


BYTE read_vic(ADDRESS addr) {

     switch (addr) {
       case 17:	/* $D011 */
	 return ((vic->reg[addr] & 0x7f) | ((CURRENT_LINE(clk) & 0x100) >> 1));

       case 18:/* $D012 raster register */
	 return (CURRENT_LINE(clk) & 0xff);

       case 19:
       case 20:/* Light Pen X,Y */
	 return (0xff);

#ifdef OLD_VIC
       case 25:/* $D019 interrupt register */
	 {
	     BYTE    tmp;
	     if (clk > rasteru)
		 update_raster(0);
	     tmp = videoint;
	     videoint = 0;
	     return (tmp);
	 }
#endif  /* OLD_VIC */

       case 26:/* $D01A interrupt control */
	 return (vic->reg[addr]);

       case 30:/* $D01E sprite sprite collision */
	 return (vic->reg[addr]);

       case 31:/* $D01F sprite background collision */
	 return (vic->reg[addr]);

#ifdef C128
       case 47:/* $D02F additional KBD rows */
	 return (vic->reg[47] | 0xf8);
#endif

       default:
	 return ((addr < 0x31) ? vic->reg[addr] : 0xff);  /* VIC IIe size */
     }
}


void store_colorram(ADDRESS addr, BYTE byte)
{
  vic->colorRam[addr] = byte;

 /* set update flag! */

}


BYTE read_colorram(ADDRESS addr)
{
  return vic->colorRam[addr];
}


/* ------------------------------------------------------------------------- */

void set_video_hooks(void)
{
}


void screen_init(void)
{
  vic = VicInit();
  VicVerticalBlank(vic);
}


void int_raster(void)
{
}


char *get_vic_bitmap()
{
  VicDrawSlab(vic, 0, 0, 320, 199);
  VicVerticalBlank(vic);
  return vic->screen;
}
